from abc import abstractmethod

from gpiozero import Device
from gpiozero.devices import GPIOMeta


class I2CDevice(Device):
    __metaclass__ = GPIOMeta

    def __init__(self, pin_factory=None) -> None:
        super().__init__(pin_factory=pin_factory)

    @abstractmethod
    def _signal_event(self, active: bool) -> None: ...

    @property
    @abstractmethod
    def bounce_time(self) -> float: ...



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from __future__ import annotations

import logging
from typing import List, Dict, Any

from .component_state import (
    L,
    P,
    log,
    SCOPE_TO_STATE_MAP,
    ComponentState,
)
from ..pdi.d4_req import D4Req
from .comp_data import CompDataMixin, CompDataHandler
from ..protocol.multibyte.multibyte_constants import TMCC2EffectsControl
from ..protocol.command_req import CommandReq
from ..protocol.command_def import CommandDefEnum
from ..pdi.constants import PdiCommand, IrdaAction, D4Action
from ..pdi.irda_req import IrdaReq
from ..protocol.constants import (
    LOCO_TYPE,
    CONTROL_TYPE,
    LOCO_TRACK_CRANE,
    TRACK_CRANE_STATE_NUMERICS,
    RPM_TYPE,
    STEAM_TYPE,
    CommandScope,
    LOCO_CLASS,
    SOUND_TYPE,
    LEGACY_CONTROL_TYPE,
)
from ..protocol.tmcc1.tmcc1_constants import TMCC1_COMMAND_TO_ALIAS_MAP, TMCC1EngineCommandEnum, TMCC1HaltCommandEnum
from ..protocol.tmcc2.tmcc2_constants import TMCC2_COMMAND_TO_ALIAS_MAP, TMCC2EngineCommandEnum
from ..protocol.tmcc1.tmcc1_constants import TMCC1EngineCommandEnum as TMCC1
from ..protocol.tmcc2.tmcc2_constants import TMCC2EngineCommandEnum as TMCC2

DIRECTIONS_SET = {
    TMCC1EngineCommandEnum.FORWARD_DIRECTION,
    TMCC2EngineCommandEnum.FORWARD_DIRECTION,
    TMCC1EngineCommandEnum.REVERSE_DIRECTION,
    TMCC2EngineCommandEnum.REVERSE_DIRECTION,
    TMCC1EngineCommandEnum.TOGGLE_DIRECTION,
    TMCC2EngineCommandEnum.TOGGLE_DIRECTION,
}
MOMENTUM_SET = {
    TMCC1EngineCommandEnum.MOMENTUM_LOW,
    TMCC1EngineCommandEnum.MOMENTUM_MEDIUM,
    TMCC1EngineCommandEnum.MOMENTUM_HIGH,
    TMCC2EngineCommandEnum.MOMENTUM_LOW,
    TMCC2EngineCommandEnum.MOMENTUM_MEDIUM,
    TMCC2EngineCommandEnum.MOMENTUM_HIGH,
    TMCC2EngineCommandEnum.MOMENTUM,
}
SPEED_SET = {
    TMCC1EngineCommandEnum.ABSOLUTE_SPEED,
    TMCC2EngineCommandEnum.ABSOLUTE_SPEED,
    (TMCC1EngineCommandEnum.ABSOLUTE_SPEED, 0),
    (TMCC2EngineCommandEnum.ABSOLUTE_SPEED, 0),
}
RPM_SET = {
    TMCC2EngineCommandEnum.DIESEL_RPM,
}
LABOR_SET = {
    TMCC2EngineCommandEnum.ENGINE_LABOR,
}
NUMERIC_SET = {
    TMCC1EngineCommandEnum.NUMERIC,
    TMCC2EngineCommandEnum.NUMERIC,
}
TRAIN_BRAKE_SET = {
    TMCC2EngineCommandEnum.TRAIN_BRAKE,
}
STARTUP_SET = {
    TMCC1EngineCommandEnum.START_UP_IMMEDIATE,
    (TMCC1EngineCommandEnum.NUMERIC, 3),
    TMCC2EngineCommandEnum.START_UP_IMMEDIATE,
    TMCC2EngineCommandEnum.START_UP_DELAYED,
}
SHUTDOWN_SET = {
    TMCC1EngineCommandEnum.SHUTDOWN_IMMEDIATE,
    (TMCC1EngineCommandEnum.NUMERIC, 5),
    TMCC2EngineCommandEnum.SHUTDOWN_DELAYED,
    (TMCC2EngineCommandEnum.NUMERIC, 5),
    TMCC2EngineCommandEnum.SHUTDOWN_IMMEDIATE,
}
ENGINE_AUX1_SET = {
    TMCC1EngineCommandEnum.AUX1_ON,
    TMCC1EngineCommandEnum.AUX1_OFF,
    TMCC1EngineCommandEnum.AUX1_OPTION_ONE,
    TMCC1EngineCommandEnum.AUX1_OPTION_TWO,
    TMCC2EngineCommandEnum.AUX1_ON,
    TMCC2EngineCommandEnum.AUX1_OFF,
    TMCC2EngineCommandEnum.AUX1_OPTION_ONE,
    TMCC2EngineCommandEnum.AUX1_OPTION_TWO,
}
ENGINE_AUX2_SET = {
    TMCC1EngineCommandEnum.AUX2_ON,
    TMCC1EngineCommandEnum.AUX2_OFF,
    TMCC1EngineCommandEnum.AUX2_OPTION_ONE,
    TMCC1EngineCommandEnum.AUX2_OPTION_TWO,
    TMCC2EngineCommandEnum.AUX2_ON,
    TMCC2EngineCommandEnum.AUX2_OFF,
    TMCC2EngineCommandEnum.AUX2_OPTION_ONE,
    TMCC2EngineCommandEnum.AUX2_OPTION_TWO,
}
SMOKE_SET = {
    TMCC1EngineCommandEnum.SMOKE_ON,
    (TMCC1EngineCommandEnum.NUMERIC, 9),
    TMCC1EngineCommandEnum.SMOKE_OFF,
    (TMCC1EngineCommandEnum.NUMERIC, 8),
    TMCC2EffectsControl.SMOKE_OFF,
    TMCC2EffectsControl.SMOKE_LOW,
    TMCC2EffectsControl.SMOKE_MEDIUM,
    TMCC2EffectsControl.SMOKE_HIGH,
}
SMOKE_LABEL = {
    TMCC1EngineCommandEnum.SMOKE_ON: "+",
    TMCC1EngineCommandEnum.SMOKE_OFF: "-",
    TMCC2EffectsControl.SMOKE_OFF: "-",
    TMCC2EffectsControl.SMOKE_LOW: "L",
    TMCC2EffectsControl.SMOKE_MEDIUM: "M",
    TMCC2EffectsControl.SMOKE_HIGH: "H",
}


class EngineState(ComponentState):
    def __init__(self, scope: CommandScope = CommandScope.ENGINE) -> None:
        if scope not in {CommandScope.ENGINE, CommandScope.TRAIN}:
            raise ValueError(f"Invalid scope: {scope}, expected ENGINE or TRAIN")
        super().__init__(scope)
        self._aux1: CommandDefEnum | None = None
        self._aux2: CommandDefEnum | None = None
        self._aux: CommandDefEnum | None = None
        self._direction: CommandDefEnum | None = None
        self._is_legacy: bool | None = None  # assume we are in TMCC mode until/unless we receive a Legacy cmd
        self._last_aux1_opt1 = None
        self._last_aux2_opt1 = None
        self._numeric: int | None = None
        self._numeric_cmd: CommandDefEnum | None = None
        self._prod_year: int | None = None
        self._start_stop: CommandDefEnum | None = None
        self._d4_rec_no: int | None = None

    def __repr__(self) -> str:
        sp = ss = name = num = mom = rl = yr = nu = lt = tb = aux = lb = sm = c = bt = ""
        if self._direction in {TMCC1EngineCommandEnum.FORWARD_DIRECTION, TMCC2EngineCommandEnum.FORWARD_DIRECTION}:
            dr = " FWD"
        elif self._direction in {TMCC1EngineCommandEnum.REVERSE_DIRECTION, TMCC2EngineCommandEnum.REVERSE_DIRECTION}:
            dr = " REV"
        else:
            dr = " N/A"

        if self.speed is not None:
            sp = f" Speed: {self.speed:03}"
            if self.speed_limit is not None:
                speed_limit = self.decode_speed_info(self.speed_limit)
                sp += f"/{speed_limit:03}"
            if self.max_speed is not None:
                max_speed = self.decode_speed_info(self.max_speed)
                sp += f"/{max_speed:03}"
        if self._start_stop is not None:
            if self._start_stop in STARTUP_SET:
                ss = " Started up"
            elif self._start_stop in SHUTDOWN_SET:
                ss = " Shut down"
        if self.momentum is not None:
            mom = f" Mom: {self.momentum_label}"
        if self.train_brake is not None:
            tb = f" TB: {self.train_brake_label}"
        if self.rpm is not None:
            rl = f" RPM: {self.rpm}"
        if self.labor is not None:
            lb = f" Labor: {self.labor:>2d}"
        if self._numeric is not None:
            nu = f" Num: {self._numeric}"
        if self.road_name is not None:
            name = f" {self.road_name}"
        if self.road_number is not None:
            num = f" #{self.road_number}"
        if self.year is not None:
            num = f" Released: {self.year}"
        if self.engine_type is not None:
            lt = f" {LOCO_TYPE.get(self.engine_type, 'NA')}"
        if self._aux2:
            aux = f" Aux2: {self._aux2.name.split('_')[-1]}"
        if self.smoke_level is not None:
            sm = f" Smoke: {self.smoke_level.name.split('_')[-1].lower():<4}"
        if self.bt_int:
            bt = f" BT: {self.bt_id}"
        ct = f" {CONTROL_TYPE.get(self.control_type, 'NA')}"
        if isinstance(self, TrainState) and self.consist_components:
            c = "\n"
            for cc in self.consist_components:
                c += f"{cc} "
        return (
            f"{self.scope.title} {self._address:04}{sp}{rl}{lb}{mom}{tb}{dr}{sm}"
            f"{name}{num}{lt}{ct}{yr}{bt}{ss}{nu}{aux}{c}"
        )

    def decode_speed_info(self, speed_info):
        if speed_info is not None and speed_info == 255:  # not set
            if self.is_legacy:
                speed_info = 195
            else:
                speed_info = 31
        return speed_info

    def update(self, command: L | P) -> None:
        from ..pdi.base_req import BaseReq

        # suppress duplicate commands that are received within 1 second; dups are common
        # in the lionel ecosystem, as commands are frequently sent twice or even 3 times
        # consecutively.
        self._is_known = True
        if command is None or (command == self._last_command and self.last_updated_ago < 1):
            return
        with self._cv:
            # in the event the initial state hasn't been loaded from the base, resubmit the command
            if self.is_comp_data_record is False:
                if isinstance(command, CommandReq):
                    from ..comm.command_listener import CommandDispatcher

                    log.info(f"Still awaiting for initial state, will retry {command}...")
                    CommandDispatcher.get().offer(command)
                    return
            super().update(command)
            if isinstance(command, CompDataMixin) and command.is_comp_data_record:
                self._update_comp_data(command.comp_data)
                if isinstance(command, D4Req):
                    self._is_legacy = True
                    self._d4_rec_no = command.record_no
            elif isinstance(command, CommandReq):
                if self._is_legacy is None:
                    self._is_legacy = command.is_tmcc2 is True or self.address > 99

                # handle some aspects of the halt command
                if command.command == TMCC1HaltCommandEnum.HALT:
                    if self.is_legacy:
                        self._aux1 = TMCC2.AUX1_OFF
                        self._aux2 = TMCC2.AUX2_OFF
                        self._aux = TMCC2.AUX2_OPTION_ONE
                    else:
                        self._aux1 = TMCC1.AUX1_OFF
                        self._aux2 = TMCC1.AUX2_OFF
                        self._aux = TMCC1.AUX2_OPTION_ONE
                    self.comp_data.speed = 0
                    self.comp_data.rpm_tmcc = 0
                    self.comp_data.labor_tmcc = 12
                    self._numeric = None
                    self._last_command = command

                # get the downstream effects of this command, as they also impact state
                cmd_effects = self.results_in(command)
                log.debug(f"Update: {command}\nEffects: {cmd_effects}")

                # handle last numeric
                if command.command in NUMERIC_SET:
                    from ..pdi.base3_buffer import Base3Buffer

                    if self.engine_type in {
                        LOCO_TRACK_CRANE,
                    }:
                        if command.data in TRACK_CRANE_STATE_NUMERICS:
                            self._numeric = command.data
                            self._numeric_cmd = command.command
                    else:
                        self._numeric = command.data
                        self._numeric_cmd = command.command
                        # numeric commands can change RPM, Volume, and reset the train
                        # force a state update for this engine/train, if we are connected
                        # to a Base 3
                        Base3Buffer.request_state_update(self.address, self.scope)
                elif cmd_effects & NUMERIC_SET:
                    numeric = self._harvest_effect(cmd_effects & NUMERIC_SET)
                    log.info(f"What to do? {command}: {numeric} {type(numeric)}")

                # Direction changes trigger several other changes; we want to avoid resettling
                # rpm, labor, and speed if the direction really didn't change
                if command.command in DIRECTIONS_SET:
                    if self._direction != command.command:
                        self._direction = self._change_direction(command.command)
                    else:
                        return
                elif cmd_effects & DIRECTIONS_SET:
                    self._direction = self._change_direction(self._harvest_effect(cmd_effects & DIRECTIONS_SET))

                # handle train brake
                if command.command in TRAIN_BRAKE_SET:
                    self.comp_data.train_brake_tmcc = command.data
                elif cmd_effects & TRAIN_BRAKE_SET:
                    self.comp_data.train_brake_tmcc = self._harvest_effect(cmd_effects & TRAIN_BRAKE_SET)

                if command.command in SMOKE_SET or (command.command, command.data) in SMOKE_SET:
                    if isinstance(command.command, TMCC2EffectsControl):
                        self.comp_data.smoke_tmcc = command.command
                    elif command.is_data and (command.command, command.data) in TMCC1_COMMAND_TO_ALIAS_MAP:
                        self.comp_data.smoke_tmcc = TMCC1_COMMAND_TO_ALIAS_MAP[(command.command, command.data)]

                # aux commands
                for cmd in {command.command} | (cmd_effects & ENGINE_AUX1_SET):
                    if cmd in ENGINE_AUX1_SET:
                        self._aux = cmd if cmd in {TMCC1.AUX1_OPTION_ONE, TMCC2.AUX1_OPTION_ONE} else self._aux
                        self._aux1 = cmd

                for cmd in {command.command} | (cmd_effects & ENGINE_AUX2_SET):
                    if cmd in ENGINE_AUX2_SET:
                        self._aux = cmd if cmd in {TMCC1.AUX2_OPTION_ONE, TMCC2.AUX2_OPTION_ONE} else self._aux
                        if cmd in {TMCC1.AUX2_OPTION_ONE, TMCC2.AUX2_OPTION_ONE}:
                            if self.time_delta(self._last_updated, self._last_aux2_opt1) > 1:
                                if self.is_legacy:
                                    self._aux2 = self.update_aux_state(
                                        self._aux2,
                                        TMCC2.AUX2_ON,
                                        TMCC2.AUX2_OPTION_ONE,
                                        TMCC2.AUX2_OFF,
                                    )
                                else:
                                    self._aux2 = self.update_aux_state(
                                        self._aux2,
                                        TMCC1.AUX2_ON,
                                        TMCC1.AUX2_OPTION_ONE,
                                        TMCC1.AUX2_OFF,
                                    )
                            self._last_aux2_opt1 = self.last_updated
                        elif cmd in {
                            TMCC1.AUX2_ON,
                            TMCC1.AUX2_OFF,
                            TMCC1.AUX2_OPTION_TWO,
                            TMCC2.AUX2_ON,
                            TMCC2.AUX2_OFF,
                            TMCC2.AUX2_OPTION_TWO,
                        }:
                            self._aux2 = cmd
                            self._last_aux2_opt1 = self.last_updated

                # handle run level/rpm
                if command.command in RPM_SET:
                    self.comp_data.rpm_tmcc = command.data
                elif cmd_effects & RPM_SET:
                    rpm = self._harvest_effect(cmd_effects & RPM_SET)
                    if isinstance(rpm, tuple) and len(rpm) == 2:
                        self.comp_data.rpm_tmcc = rpm[1]
                    elif isinstance(rpm, CommandDefEnum):
                        if log.isEnabledFor(logging.DEBUG):
                            log.debug(f"{command} {rpm} {type(rpm)} {rpm.command_def} {type(rpm.command_def)}")
                        self.comp_data.rpm_tmcc = 0
                    else:
                        if log.isEnabledFor(logging.DEBUG):
                            log.debug(f"{command} {rpm} {type(rpm)} {cmd_effects}")
                        self.comp_data.rpm_tmcc = 0

                # handle labor
                if command.command in LABOR_SET:
                    self.comp_data.labor_tmcc = command.data
                elif cmd_effects & LABOR_SET:
                    labor = self._harvest_effect(cmd_effects & LABOR_SET)
                    if isinstance(labor, tuple) and len(labor) == 2:
                        self.comp_data.labor_tmcc = labor[1]
                    else:
                        if log.isEnabledFor(logging.DEBUG):
                            log.debug(f"{command} {labor} {type(labor)} {cmd_effects}")
                        self.comp_data.speed = 0

                # handle speed
                if command.command in SPEED_SET:
                    self.comp_data.speed = command.data
                elif cmd_effects & SPEED_SET:
                    speed = self._harvest_effect(cmd_effects & SPEED_SET)
                    if isinstance(speed, tuple) and len(speed) > 1:
                        self.comp_data.speed = speed[1]
                    else:
                        if log.isEnabledFor(logging.DEBUG):
                            log.debug(f"{command} {speed} {type(speed)} {cmd_effects}")
                        self.comp_data.speed = 0

                # handle momentum
                if command.command in MOMENTUM_SET:
                    if command.command in {
                        TMCC1EngineCommandEnum.MOMENTUM_LOW,
                        TMCC2EngineCommandEnum.MOMENTUM_LOW,
                    }:
                        self.comp_data.momentum_tmcc = 0
                    if command.command in {
                        TMCC1EngineCommandEnum.MOMENTUM_MEDIUM,
                        TMCC2EngineCommandEnum.MOMENTUM_MEDIUM,
                    }:
                        self.comp_data.momentum_tmcc = 3
                    if command.command in {
                        TMCC1EngineCommandEnum.MOMENTUM_HIGH,
                        TMCC2EngineCommandEnum.MOMENTUM_HIGH,
                    }:
                        self.comp_data.momentum_tmcc = 7
                    elif command.command == TMCC2EngineCommandEnum.MOMENTUM:
                        self.comp_data.momentum_tmcc = command.data

                # handle startup/shutdown
                if command.command in STARTUP_SET:
                    self._start_stop = command.command
                elif command.command in SHUTDOWN_SET:
                    self._start_stop = command.command
                elif cmd_effects & STARTUP_SET:
                    startup = self._harvest_effect(cmd_effects & STARTUP_SET)
                    if isinstance(startup, CommandDefEnum):
                        self._start_stop = startup
                    elif command.is_data and (command.command, command.data) in TMCC2_COMMAND_TO_ALIAS_MAP:
                        self._start_stop = TMCC2_COMMAND_TO_ALIAS_MAP[(command.command, command.data)]
                    elif command.is_data and (command.command, command.data) in TMCC1_COMMAND_TO_ALIAS_MAP:
                        self._start_stop = TMCC1_COMMAND_TO_ALIAS_MAP[(command.command, command.data)]
                        self._numeric = command.data
                elif cmd_effects & SHUTDOWN_SET:
                    shutdown = self._harvest_effect(cmd_effects & SHUTDOWN_SET)
                    if isinstance(shutdown, CommandDefEnum):
                        self._start_stop = shutdown
                    elif command.is_data and (command.command, command.data) in TMCC2_COMMAND_TO_ALIAS_MAP:
                        self._start_stop = TMCC2_COMMAND_TO_ALIAS_MAP[(command.command, command.data)]
                    elif command.is_data and (command.command, command.data) in TMCC1_COMMAND_TO_ALIAS_MAP:
                        self._start_stop = TMCC1_COMMAND_TO_ALIAS_MAP[(command.command, command.data)]
                        self._numeric = command.data
            elif (
                isinstance(command, BaseReq)
                and command.status == 0
                and command.pdi_command
                in {
                    PdiCommand.UPDATE_ENGINE_SPEED,
                    PdiCommand.UPDATE_TRAIN_SPEED,
                }
            ):
                from ..pdi.base_req import EngineBits

                if self.speed is None and command.is_valid(EngineBits.SPEED):
                    self.comp_data.speed = command.speed
            elif isinstance(command, BaseReq) and command.pdi_command == PdiCommand.BASE_MEMORY and command.data_bytes:
                from .comp_data import BASE_MEMORY_ENGINE_READ_MAP

                tpl = BASE_MEMORY_ENGINE_READ_MAP.get(command.start, None)
                if isinstance(tpl, CompDataHandler):
                    setattr(self, tpl.field, tpl.from_bytes(command.data_bytes))
            elif isinstance(command, IrdaReq) and command.action == IrdaAction.DATA:
                self._prod_year = command.year
            elif isinstance(command, D4Req):
                if command.action == D4Action.MAP:
                    if command.record_no == 0xFFFF:  # delete record
                        # TODO: delete state record
                        pass
                    elif command.record_no is not None:
                        self._d4_rec_no = command.record_no
            self.changed.set()
            self._cv.notify_all()

    def _change_direction(self, new_dir: CommandDefEnum) -> CommandDefEnum:
        if new_dir in {TMCC1EngineCommandEnum.TOGGLE_DIRECTION, TMCC2EngineCommandEnum.TOGGLE_DIRECTION}:
            if self.direction is not None:
                if self.is_legacy is True and self.direction in {
                    TMCC2EngineCommandEnum.FORWARD_DIRECTION,
                    TMCC2EngineCommandEnum.REVERSE_DIRECTION,
                }:
                    new_dir = (
                        TMCC2EngineCommandEnum.FORWARD_DIRECTION
                        if self.direction == TMCC2EngineCommandEnum.REVERSE_DIRECTION
                        else TMCC2EngineCommandEnum.REVERSE_DIRECTION
                    )
                elif self.is_tmcc is True and self.direction in {
                    TMCC1EngineCommandEnum.FORWARD_DIRECTION,
                    TMCC1EngineCommandEnum.REVERSE_DIRECTION,
                }:
                    new_dir = (
                        TMCC1EngineCommandEnum.FORWARD_DIRECTION
                        if self.direction == TMCC1EngineCommandEnum.REVERSE_DIRECTION
                        else TMCC1EngineCommandEnum.REVERSE_DIRECTION
                    )
                else:
                    new_dir = None
        return new_dir

    def as_bytes(self) -> list[bytes]:
        from ..pdi.base_req import BaseReq

        packets = []
        # encode name, number, momentum, speed, and rpm using PDI command
        if self.tmcc_id <= 99:
            pdi = BaseReq(self.address, PdiCommand.BASE_MEMORY, scope=self.scope, state=self)
        else:
            pdi_cmd = PdiCommand.D4_ENGINE if self.scope == CommandScope.ENGINE else PdiCommand.D4_TRAIN
            pdi = D4Req(self.record_no, pdi_cmd, state=self)
        packets.append(pdi.as_bytes)
        if self._start_stop is not None:
            packets.append(CommandReq.build(self._start_stop, self.address, scope=self.scope).as_bytes)
        if self.smoke_level is not None:
            packets.append(CommandReq.build(self.smoke_level, self.address, scope=self.scope).as_bytes)
        if self._direction is not None:
            # the direction state will have encoded in it the syntax (tmcc1 or tmcc2)
            packets.append(CommandReq.build(self._direction, self.address, scope=self.scope).as_bytes)
        if self._numeric is not None and self._numeric_cmd is not None:
            if self.engine_type in {
                LOCO_TRACK_CRANE,
            }:
                packets.append(
                    CommandReq.build(
                        self._numeric_cmd,
                        self.address,
                        data=self._numeric,
                        scope=self.scope,
                    ).as_bytes
                )
        if self._aux is not None:
            packets.append(CommandReq.build(self._aux, self.address).as_bytes)
        if self._aux1 is not None:
            packets.append(CommandReq.build(self.aux1, self.address).as_bytes)
        if self._aux2 is not None:
            packets.append(CommandReq.build(self.aux2, self.address).as_bytes)
        return packets

    @property
    def is_rpm(self) -> bool:
        return self.comp_data.engine_type in RPM_TYPE

    @property
    def is_steam(self) -> bool:
        return self.comp_data.engine_type in STEAM_TYPE

    @property
    def speed(self) -> int:
        return self.comp_data.speed

    @property
    def speed_limit(self) -> int:
        return self.comp_data.speed_limit

    @property
    def max_speed(self) -> int:
        return self.comp_data.max_speed

    @property
    def speed_max(self) -> int | None:
        if self.max_speed and self.max_speed != 255 and self.speed_limit != 255:
            ms = min(self.max_speed, self.speed_limit)
        elif self.speed_limit and self.speed_limit != 255:
            ms = self.speed_limit
        elif self.max_speed and self.max_speed != 255:
            ms = self.max_speed
        else:
            ms = 199 if self.is_legacy else 31
        if self.is_legacy is False and ms > 31:
            ms = 31
        return ms

    @property
    def speed_label(self) -> str:
        return self._as_label(self.speed)

    @property
    def bt_int(self) -> int:
        return self.comp_data.bt_id

    # noinspection PyTypeChecker
    @property
    def bt_id(self) -> str:
        if self.bt_int:
            return int.to_bytes(self.bt_int, 2, "big").hex().upper()
        return None

    @property
    def numeric(self) -> int:
        return self._numeric

    @property
    def momentum(self) -> int:
        return self.comp_data.momentum_tmcc

    @property
    def momentum_label(self) -> str:
        return self._as_label(self.momentum)

    @property
    def fuel_level(self) -> int:
        return self.comp_data.fuel_level

    @property
    def fuel_level_pct(self) -> int:
        if self.fuel_level is not None:
            return int(round((self.fuel_level / 255.0) * 100.0))
        return self.fuel_level

    @property
    def fuel_level_label(self) -> str:
        return self._as_label(self.fuel_level)

    @property
    def water_level(self) -> int:
        return self.comp_data.water_level

    @property
    def water_level_pct(self) -> int:
        if self.water_level is not None:
            return int(round((self.water_level / 255.0) * 100.0))
        return self.water_level

    @property
    def water_level_label(self) -> str:
        return self._as_label(self.water_level)

    @property
    def rpm(self) -> int:
        return self.comp_data.rpm_tmcc if self.is_rpm else 0

    @property
    def rpm_label(self) -> str:
        return self._as_label(self.rpm)

    @property
    def labor(self) -> int:
        return self.comp_data.labor_tmcc

    @property
    def labor_label(self) -> str:
        return self._as_label(self.labor)

    @property
    def smoke_level(self) -> CommandDefEnum:
        return self.comp_data.smoke_tmcc

    @property
    def smoke_label(self) -> str:
        return SMOKE_LABEL.get(self.smoke_level, None)

    @property
    def train_brake(self) -> int:
        return self.comp_data.train_brake_tmcc

    @property
    def train_brake_label(self) -> str:
        return self._as_label(self.train_brake)

    @property
    def control_type(self) -> int:
        return self.comp_data.control_type

    @property
    def control_type_label(self) -> str:
        return CONTROL_TYPE.get(self.control_type, "NA")

    @property
    def sound_type(self) -> int:
        return self.comp_data.sound_type

    @property
    def sound_type_label(self) -> str:
        return SOUND_TYPE.get(self.sound_type, "NA")

    @property
    def engine_type(self) -> int:
        return self.comp_data.engine_type

    @property
    def engine_type_label(self) -> str:
        return LOCO_TYPE.get(self.engine_type, "NA")

    @property
    def engine_class(self) -> int:
        return self.comp_data.engine_class

    @property
    def engine_class_label(self) -> str:
        return LOCO_CLASS.get(self.engine_class, "NA")

    @property
    def direction(self) -> CommandDefEnum | None:
        return self._direction

    @property
    def direction_label(self) -> str:
        dr = "--"
        if self._direction in {TMCC1EngineCommandEnum.FORWARD_DIRECTION, TMCC2EngineCommandEnum.FORWARD_DIRECTION}:
            dr = "FW"
        elif self._direction in {TMCC1EngineCommandEnum.REVERSE_DIRECTION, TMCC2EngineCommandEnum.REVERSE_DIRECTION}:
            dr = "RV"
        return dr

    @property
    def stop_start(self) -> CommandDefEnum | None:
        return self._start_stop

    @property
    def is_started(self) -> bool:
        return self._start_stop in STARTUP_SET

    @property
    def is_shutdown(self) -> bool:
        return self._start_stop in SHUTDOWN_SET

    @property
    def year(self) -> int:
        return self._prod_year

    @property
    def is_aux_on(self) -> bool:
        return self._aux in {TMCC1.AUX1_OPTION_ONE, TMCC2.AUX1_OPTION_ONE}

    @property
    def is_aux_off(self) -> bool:
        return self.is_aux_on is False

    @property
    def aux1(self) -> CommandDefEnum:
        return self._aux1

    @property
    def aux2(self) -> CommandDefEnum:
        return self._aux2

    @property
    def is_aux1(self) -> bool:
        return self._aux2 in {TMCC1.AUX1_ON, TMCC2.AUX1_ON}

    @property
    def is_aux2(self) -> bool:
        return self._aux2 in {TMCC1.AUX2_ON, TMCC2.AUX2_ON}

    @property
    def record_no(self) -> int:
        return self._d4_rec_no

    @property
    def is_tmcc(self) -> bool:
        return self._is_legacy is False

    @property
    def is_legacy(self) -> bool:
        if self._is_legacy is None:
            if self.scope == CommandScope.ENGINE and self.address > 99:
                self._is_legacy = True
            elif self.control_type is not None:
                if self.control_type == LEGACY_CONTROL_TYPE:
                    self._is_legacy = True
                else:
                    self._is_legacy = False
        return self._is_legacy is True

    @property
    def is_lcs(self) -> bool:
        return False

    def as_dict(self) -> Dict[str, Any]:
        d = super()._as_dict()
        for elem in ["speed", "speed_limit", "max_speed", "train_brake", "momentum", "rpm", "labor", "year"]:
            if hasattr(self, elem):
                val = getattr(self, elem)
                d[elem] = val if val is not None and val != 255 else None
        d["direction"] = self.direction.name.lower() if self.direction else None
        d["smoke"] = self.smoke_level.name.lower() if self.smoke_level else None
        d["control"] = self.control_type_label.lower() if self.control_type is not None else None
        d["sound_type"] = self.sound_type_label.lower() if self.sound_type is not None else None
        d["engine_type"] = self.engine_type_label.lower() if self.engine_type is not None else None
        d["engine_class"] = self.engine_class_label.lower() if self.engine_class is not None else None
        if isinstance(self, TrainState):
            d["flags"] = self.consist_flags
            d["components"] = {c.tmcc_id: c.info for c in self.consist_components}
        return d


class TrainState(EngineState):
    from ..pdi.base3_component import ConsistComponent

    def __init__(self, scope: CommandScope = CommandScope.TRAIN) -> None:
        if scope != CommandScope.TRAIN:
            raise ValueError(f"Invalid scope: {scope}, expected {CommandScope.TRAIN.name}")
        super().__init__(scope)
        # hard code TMCC2, for now
        self._is_legacy = True

    @property
    def consist_flags(self) -> int:
        return self.comp_data.consist_flags

    @property
    def consist_components(self) -> List[ConsistComponent]:
        return self.comp_data.consist_comps


SCOPE_TO_STATE_MAP.update({CommandScope.ENGINE: EngineState})
SCOPE_TO_STATE_MAP.update({CommandScope.TRAIN: TrainState})



import sys
from typing import List

from smbus2 import SMBus

if sys.platform == "linux":
    try:
        DEFAULT_SMBUS = SMBus(1)
    except FileNotFoundError:
        DEFAULT_SMBUS = None
else:
    DEFAULT_SMBUS = None


class I2C:
    def __init__(self, smbus: SMBus = DEFAULT_SMBUS):
        """
        Wrapper class for a smbus
        :param smbus: the smbus to send and receive data from smbus.SMBus(1)
        """
        self.bus: SMBus = smbus

    def write_to(self, address: int, offset: int, value: int) -> None:
        self.bus.write_byte_data(address, offset, value)

    def read_from(self, address: int, offset: int = 0) -> int:
        return self.bus.read_byte_data(address, offset)

    def read(self, address: int) -> int:
        return self.bus.read_byte(address)

    def scan(self) -> List[int]:
        devices = list()
        for address in range(255):
            try:
                self.bus.read_byte(address)  # try to read byte
                devices.append(address)
            except KeyError:  # exception if read_byte fails
                pass
        return devices



#!/usr/bin/env python3
#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
import logging
from argparse import ArgumentParser
from typing import List

from . import CliBase
from ..pdi.bpc2_req import Bpc2Req
from ..pdi.constants import PdiCommand, Bpc2Action
from ..protocol.command_base import CommandBase
from ..protocol.constants import CommandScope, PROGRAM_NAME
from ..utils.argument_parser import PyTrainArgumentParser

log = logging.getLogger(__name__)


class Bpc2Cmd(CommandBase):
    def __init__(self, tmcc_id: int, req: Bpc2Req, server: str = None) -> None:
        if tmcc_id < 1 or tmcc_id > 99:
            raise ValueError("Bpc2 ID must be between 1 and 99")
        super().__init__(None, req, tmcc_id, CommandScope.BPC2, server=server)
        self._command = self._build_command()

    def _build_command(self) -> bytes | None:
        return self.command_req.as_bytes

    def _command_prefix(self) -> bytes | None:
        pass

    def _encode_address(self, command_op: int) -> bytes | None:
        pass


class Bpc2Cli(CliBase):
    """
    Issue Bpc2 Commands.
    """

    @classmethod
    def command_parser(cls) -> ArgumentParser:
        parser = PyTrainArgumentParser(add_help=False)
        parser.add_argument("bpc2", metavar="Bbc2 TMCC ID", type=int, help="Bpc2 to fire")

        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "-on",
            action="store_true",
            help="Turn block on",
        )
        group.add_argument(
            "-off",
            action="store_true",
            help="Turn block off",
        )

        parser.add_argument(
            "-server",
            action="store",
            help=f"IP Address of {PROGRAM_NAME} server, if client. Server communicates with Base 3/LCS SER2",
        )
        # fire command
        return PyTrainArgumentParser("Operate specified Bpc2 Track power district (1 - 99)", parents=[parser])

    def __init__(self, arg_parser: ArgumentParser = None, cmd_line: List[str] = None, do_fire: bool = True) -> None:
        super().__init__(arg_parser, cmd_line, do_fire)
        self._bpc2 = self._args.bpc2
        self._state = 1 if self._args.on is True else 0
        req = Bpc2Req(self._bpc2, PdiCommand.BPC2_SET, Bpc2Action.CONTROL3, state=self._state)

        try:
            cmd = Bpc2Cmd(self._bpc2, req, server=self._server)
            if self.do_fire:
                cmd.fire(server=self._server)
            self._command = cmd
        except ValueError as ve:
            log.exception(ve)



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

import re

U33C_PATTERN = re.compile(r"[A-Z]\d{2}[A-Z]")


def title(text: str):
    if text:
        parts = text.strip().split(" ")
        for i, part in enumerate(parts):
            part = part.strip().upper()
            if len(part) > 3:
                if part.startswith("SD") and (part.endswith("ACE") or part.endswith("AC")):
                    if part.endswith("ACE"):
                        part = part.replace("ACE", "ACe")
                    else:
                        part = part.replace("AC", "ACe")
                elif part.startswith("SD"):
                    if i + 1 < len(parts) and parts[i + 1].upper() == "ACE":
                        part = part + "ACe"
                        part = part.replace("SD-", "SD")
                        parts[i + 1] = ""
                elif part.startswith("FA-") or part.startswith("RS-") or part.startswith("GP"):
                    pass
                elif U33C_PATTERN.match(part):
                    pass
                else:
                    part = part.capitalize()
            elif part in {"NEW", "OLD", "CAR", "RIO", "PAD", "BEE", "TO"}:
                part = part.capitalize()
            parts[i] = part
        text = " ".join([p for p in parts if p])
    return text



from threading import Condition, RLock


class TextBuffer:
    def __init__(self, rows: int = 4, cols: int = 0, auto_update: bool = True) -> None:
        super().__init__()
        self._rows = rows
        self._cols = cols
        self._cursor_pos = (0, 0)
        self._buffer: list[str] = list()
        self._cv = Condition(RLock())
        self._auto_update = auto_update
        self._changed_rows: set[int] = set()

    def __repr__(self) -> str:
        with self._cv:
            return "\n".join(self._buffer)

    def __len__(self) -> int:
        with self._cv:
            return len(self._buffer)

    def __getitem__(self, index: int):
        with self._cv:
            if 0 <= index < len(self._buffer):
                return self._buffer[index]
            else:
                raise IndexError(f"Index {index} out of range")

    def __setitem__(self, index: int, value: str):
        with self._cv:
            if self.rows > index >= len(self._buffer):
                for _ in range(index - len(self._buffer) + 1):
                    self._buffer.append("")
            if 0 <= index < self.rows:
                if self._buffer[index] != value:
                    self._buffer[index] = value
                    self._changed_rows.add(index)
                    self._cursor_pos = (index, len(value))
                    self.__do_notify()
            else:
                raise IndexError(f"Index {index} out of range")

    def __delitem__(self, index: int | slice):
        with self._cv:
            del self._buffer[index]
            self._changed_rows = set(range(self.rows))
            self.__do_notify()

    def __iter__(self):
        return iter(self._buffer)

    def insert(self, index, item):
        with self._cv:
            if index >= self.rows or index < 0 or (len(self) + 1) > self.rows:
                raise IndexError(f"Index {index} out of range")
            self._buffer.insert(index, item)
            self._changed_rows = set(range(index, self.rows))
            self.__do_notify()

    def index(self, item, start=0, end=None):
        with self._cv:
            return self._buffer.index(item, start, end if end is not None else len(self._buffer))

    def count(self, item: str):
        with self._cv:
            return self._buffer.count(item)

    def __contains__(self, item: str):
        with self._cv:
            return item in self._buffer

    def reverse(self):
        with self._cv:
            self._buffer.reverse()
            self._changed_rows = set(range(len(self._buffer)))
            self.__do_notify()

    def append(self, item: str):
        self.add(item)

    def copy(self):
        raise NotImplementedError

    def pop(self, index=-1):
        raise NotImplementedError

    def extend(self, iterable):
        raise NotImplementedError

    @property
    def auto_update(self) -> bool:
        return self._auto_update

    @auto_update.setter
    def auto_update(self, value: bool) -> None:
        self._auto_update = value

    @property
    def rows(self) -> int:
        return self._rows

    @property
    def cols(self) -> int:
        return self._cols

    @property
    def synchronizer(self) -> Condition:
        return self._cv

    @property
    def cursor_pos(self) -> tuple[int, int]:
        return self._cursor_pos

    @cursor_pos.setter
    def cursor_pos(self, pos: tuple[int, int] | int, col: int = None) -> None:
        with self.synchronizer:
            if isinstance(pos, int) and isinstance(col, int):
                pos = (pos, col)
            self._vet_cursor_pos(pos)
            self._cursor_pos = pos

    def _vet_cursor_pos(self, pos):
        if isinstance(pos, tuple) and len(pos) == 2:
            if pos[0] < 0 or pos[0] >= self.rows:
                raise AttributeError(f"Invalid row position: {pos[0]}")
            if self.cols and (pos[1] < 0 or pos[1] >= self.cols):
                raise AttributeError(f"Invalid column position: {pos[1]}")
        else:
            raise AttributeError(f"Invalid cursor position: {pos}")

    @property
    def row(self) -> int:
        return self._cursor_pos[0]

    @property
    def col(self) -> int:
        return self._cursor_pos[1]

    @property
    def buffer(self) -> list[str]:
        return self._buffer.copy()

    @property
    def is_dirty(self) -> bool:
        return len(self._changed_rows) > 0

    @property
    def changed_rows(self) -> list[int]:
        with self._cv:
            changes = sorted(self._changed_rows)
            self._changed_rows.clear()
            return changes

    def clear(self, notify: bool = True) -> None:
        with self._cv:
            for i, row in enumerate(self._buffer):
                if row:
                    self._changed_rows.add(i)
            self._buffer.clear()
            self._cursor_pos = (0, 0)
            if notify is True:
                self.__do_notify()

    def add(self, row: str) -> bool:
        with self._cv:
            if len(self._buffer) < self.rows:
                self._buffer.append(row)
                row_no = len(self._buffer) - 1
                self._cursor_pos = (row_no, len(row))
                self._changed_rows.add(row_no)
                self.__do_notify()
                return True
            return False

    def write_chr(self, int_chr: int, at: tuple[int, int] = None) -> None:
        if isinstance(int_chr, int) and 0 <= int_chr <= 255:
            self.write(chr(int_chr), at=at)

    def write(
        self,
        c: int | str,
        at: tuple[int, int] | int = None,
        fmt: str = None,
        center: bool = False,
    ) -> None:
        with self._cv:
            if isinstance(at, int):
                at = (at, 0)
            elif isinstance(at, tuple) and len(at) == 2:
                pass
            else:
                at = self._cursor_pos
            self._vet_cursor_pos(at)

            if len(self._buffer) <= at[0]:
                for _ in range(at[0] + 1 - len(self._buffer)):
                    self._buffer.append("")

            if fmt is not None:
                fmt = f"{fmt[1:] if fmt.startswith(':') else fmt}"

            if isinstance(c, int):
                if fmt:
                    s = f"{c:{fmt}}"
                    fmt = None
                else:
                    s = str(c)
            else:
                s = str(c)
            if fmt is not None:
                s = f"{s:{fmt}}"
            # append new data to buffer
            row = orig_row = self[at[0]]
            if center is True:
                pad_chrs = int((self.cols - len(s)) / 2)
                row = (" " * pad_chrs if pad_chrs > 0 else "") + s
                col = len(row)
            elif len(row) <= at[1]:
                # pad row with spaces, if cursor pos is > current row length
                row += " " * (at[1] - len(self[at[0]])) + s
                col = len(row)
            elif at[1] + len(s) > len(row):
                # overwrite portion of the row
                row = row[: at[1]] + s
                col = len(row)
            else:
                # row = row[: at[1]] + s + row[at[1] + len(s) :]
                remainder = row[at[1] + len(s) :]
                row = row[: at[1]] + s
                col = len(row)
                row += remainder
            # the cursor position might have changed even
            # if the string is the same
            self._cursor_pos = (at[0], col)
            # if the row contents didn't change, don't do any work
            if row != orig_row:
                self._buffer[at[0]] = row  # calling setitem changes cursor
                self._changed_rows.add(at[0])
                self.__do_notify()

    def __do_notify(self) -> None:
        """
        Notify all waiting threads that the buffer has changed.
        Must be called from within the synchronizer's lock.
        """
        if self._auto_update is True:
            self._cv.notify_all()



import re

# noinspection PyPackageRequirements
import pytest

from src.pytrain.utils.validations import Validations
from test.test_base import TestBase


class TestValidations(TestBase):
    def test_validate_int(self):
        # assert valid integers do not generate exceptions
        assert Validations.validate_int(5) == 5

        # test floats are converted
        assert Validations.validate_int(5.4) == 5  # type: ignore
        assert Validations.validate_int(5.99) == 5  # type: ignore

        # test numeric strings are converted
        assert Validations.validate_int("5") == 5  # type: ignore

        # test non-numeric strings  throw ValueError
        v = "text"
        with pytest.raises(ValueError, match="'text' is not an integer"):
            assert Validations.validate_int(v) == 10  # type: ignore

        # test None throw TypeError
        v = None
        with pytest.raises(TypeError, match="'None' is not an integer"):
            assert Validations.validate_int(v) == 10  # type: ignore

        # test minimum value
        v = 5
        with pytest.raises(ValueError, match=f"'{v}' must be equal to or greater than {10}"):
            assert Validations.validate_int(5, min_value=10) == 10

        # test maximum value
        v = 15
        with pytest.raises(ValueError, match=f"'{v}' must be less than or equal to {10}"):
            assert Validations.validate_int(15, max_value=10) == 10

        # test label
        with pytest.raises(ValueError, match=re.escape("Value must be less than or equal to 10 (15)")):
            assert Validations.validate_int(15, max_value=10, label="Value") == 10



import atexit
from threading import Thread, Condition, RLock
from time import time

from guizero import App, PushButton, Text, Box

from ..comm.command_listener import CommandDispatcher
from ..db.state_watcher import StateWatcher
from ..protocol.constants import CommandScope
from ..db.component_state_store import ComponentStateStore
from ..protocol.command_req import CommandReq
from ..protocol.tmcc1.tmcc1_constants import TMCC1EngineCommandEnum

from ..gpio.gpio_handler import GpioHandler
from ..utils.path_utils import find_file


class LaunchGui(Thread):
    def __init__(self, tmcc_id: int = 39):
        # initialize guizero thread
        super().__init__(daemon=True, name=f"Pad {tmcc_id} GUI")
        self.tmcc_id = tmcc_id
        self._cv = Condition(RLock())

        self.launch_jpg = find_file("launch.jpg")
        self.abort_jpg = find_file("abort.jpg")
        self.siren_on = find_file("red_light.jpg")
        self.siren_off = find_file("red_light_off.jpg")
        self.on_button = find_file("on_button.jpg")
        self.off_button = find_file("off_button.jpg")
        self.left_arrow = find_file("left_arrow.jpg")
        self.right_arrow = find_file("right_arrow.jpg")

        self.counter = None

        self.app = self.upper_box = self.lower_box = self.message = None
        self.launch = self.abort = self.pad = self.count = self.label = None
        self.gantry_box = self.siren_box = self.klaxon_box = self.lights_box = None
        self.power_button = self.lights_button = self.siren_button = self.klaxon_button = None
        self.gantry_rev = self.gantry_fwd = None

        self.power_on_req = CommandReq(TMCC1EngineCommandEnum.START_UP_IMMEDIATE, tmcc_id)
        self.power_off_req = CommandReq(TMCC1EngineCommandEnum.SHUTDOWN_IMMEDIATE, tmcc_id)
        self.reset_req = CommandReq(TMCC1EngineCommandEnum.NUMERIC, tmcc_id, 0)
        self.launch_now_req = CommandReq(TMCC1EngineCommandEnum.FRONT_COUPLER, tmcc_id)
        self.abort_now_req = CommandReq(TMCC1EngineCommandEnum.NUMERIC, tmcc_id, 5)
        self.gantry_rev_req = CommandReq(TMCC1EngineCommandEnum.NUMERIC, tmcc_id, 6)
        self.gantry_fwd_req = CommandReq(TMCC1EngineCommandEnum.NUMERIC, tmcc_id, 3)
        self.lights_on_req = CommandReq(TMCC1EngineCommandEnum.AUX2_ON, tmcc_id)
        self.lights_off_req = CommandReq(TMCC1EngineCommandEnum.AUX2_OFF, tmcc_id)
        self.siren_req = CommandReq(TMCC1EngineCommandEnum.BLOW_HORN_ONE, tmcc_id)
        self.klaxon_req = CommandReq(TMCC1EngineCommandEnum.RING_BELL, tmcc_id)
        self.launch_15_req = CommandReq(TMCC1EngineCommandEnum.REAR_COUPLER, tmcc_id)
        self.launch_seq_act = CommandReq(TMCC1EngineCommandEnum.AUX1_OPTION_ONE, tmcc_id).as_action(duration=3.5)

        # listen for state changes
        self._dispatcher = CommandDispatcher.get()
        self._state_store = ComponentStateStore.get()
        self._synchronized = False
        self._sync_state = self._state_store.get_state(CommandScope.SYNC, 99)
        self._monitored_state = None
        self._last_cmd = None
        self._last_cmd_at = 0
        self._launch_seq_time_trigger = None
        self._is_countdown = False
        self.started_up = False
        if self._sync_state and self._sync_state.is_synchronized is True:
            self._sync_watcher = None
            self.on_sync()
        else:
            self._sync_watcher = StateWatcher(self._sync_state, self.on_sync)
        atexit.register(self.close)

    def on_sync(self) -> None:
        if self._sync_state.is_synchronized:
            if self._sync_watcher:
                self._sync_watcher.shutdown()
                self._sync_watcher = None
            self._synchronized = True
            self._monitored_state = self._state_store.get_state(CommandScope.ENGINE, self.tmcc_id, False)
            if self._monitored_state is None:
                raise ValueError(f"No state found for tmcc_id: {self.tmcc_id}")
            # start GUI
            self.start()
            # listen for state updates
            self._dispatcher.subscribe(self, CommandScope.ENGINE, self.tmcc_id)

    def close(self) -> None:
        pass

    def sync_gui_state(self) -> None:
        if self._monitored_state:
            # power on?
            if self._monitored_state.is_started is True:
                self.do_power_on()
                self.lights_on_req.send()
            else:
                self.do_power_off()
                self.lights_off_req.send()
            # lights on?
            if self._monitored_state.is_aux2 is True:
                self.do_lights_on()
            else:
                self.do_lights_off()

    def __call__(self, cmd: CommandReq) -> None:
        with self._cv:
            # handle launch sequence differently
            if cmd.command == TMCC1EngineCommandEnum.AUX1_OPTION_ONE:
                if self._launch_seq_time_trigger is None:
                    if self._last_cmd != cmd or (time() - self._last_cmd_at) > 5:
                        self._launch_seq_time_trigger = time()
                else:
                    if self._last_cmd == cmd and (time() - self._launch_seq_time_trigger) > 3.1:
                        if self._is_countdown is False:
                            self.do_launch(76, detected=True)
                        self._launch_seq_time_trigger = None
                self._last_cmd = cmd
                self._last_cmd_at = time()
                return
            else:
                self._launch_seq_time_trigger = None
            if cmd != self._last_cmd or (time() - self._last_cmd_at) >= 1.0:
                self._last_cmd_at = time()
                if cmd.command == TMCC1EngineCommandEnum.NUMERIC:
                    if cmd.data in (3, 6):
                        # mark launch pad as on and lights as on
                        self.do_power_on()
                        self.do_lights_on()
                    elif cmd.data == 5:
                        self.do_lights_off()
                        self.do_power_off()
                    elif cmd.data == 0:  # reset
                        self.do_klaxon_off()
                        if self._is_countdown is True:
                            self.do_abort(detected=True)
                elif self.is_active is True:
                    if cmd.command == TMCC1EngineCommandEnum.REAR_COUPLER:
                        self.do_power_on()
                        self.do_launch(15, detected=True, hold=False)
                    elif cmd.command == TMCC1EngineCommandEnum.AUX2_OPTION_ONE:
                        if self._monitored_state.is_aux2 is True:
                            self.do_lights_on()
                        else:
                            self.do_lights_off()
                    elif cmd.command == TMCC1EngineCommandEnum.AUX2_ON:
                        self.do_lights_on()
                    elif cmd.command == TMCC1EngineCommandEnum.AUX2_OFF:
                        self.do_lights_off()

            # remember last command
            self._last_cmd = cmd

    @property
    def is_active(self) -> bool:
        return True if self._monitored_state and self._monitored_state.is_started is True else False

    def run(self):
        GpioHandler.cache_handler(self)
        self.app = app = App(title="Launch Pad", width=480, height=320)
        app.full_screen = True
        self.upper_box = upper_box = Box(app, layout="grid", border=False)

        self.launch = PushButton(
            upper_box,
            image=self.launch_jpg,
            height=128,
            width=128,
            grid=[0, 0, 1, 2],
            align="left",
            command=self.do_launch,
        )

        self.abort = PushButton(
            upper_box,
            image=self.abort_jpg,
            height=128,
            width=128,
            grid=[4, 0, 1, 2],
            align="right",
            command=self.do_abort,
        )

        if self.tmcc_id == 39:
            self.pad = Text(upper_box, text="Pad 39A", grid=[1, 0, 2, 1], size=30, bold=True)
        else:
            self.pad = Text(upper_box, text=f"Pad {self.tmcc_id}", grid=[1, 0, 2, 1], size=28)

        countdown_box = Box(upper_box, layout="auto", border=True, grid=[1, 1, 2, 1])
        self.label = Text(
            countdown_box,
            text="T-Minus",
            align="left",
            size=16,
            height=2,
            bg="black",
            color="white",
            italic=True,
        )
        self.count = Text(
            countdown_box,
            text="-00:00",
            align="right",
            size=18,
            height=2,
            font="DigitalDream",
            bg="black",
            color="white",
            italic=True,
        )

        _ = Text(upper_box, text=" ", grid=[0, 2, 5, 1], size=10)
        self.message = Text(upper_box, text="", grid=[0, 3, 5, 1], size=24, color="red", bold=True, align="top")

        self.lower_box = lower_box = Box(app, border=2, align="bottom")
        power_box = Box(lower_box, layout="grid", border=2, align="left")
        _ = Text(power_box, text="Power", grid=[0, 0], size=16, underline=True)
        self.power_button = PushButton(
            power_box,
            image=self.on_button,
            grid=[0, 1],
            command=self.toggle_power,
            height=72,
            width=72,
        )

        self.lights_box = lights_box = Box(lower_box, layout="grid", border=2, align="left")
        _ = Text(lights_box, text="Lights", grid=[0, 0], size=16, underline=True)
        self.lights_button = PushButton(
            lights_box,
            image=self.on_button,
            grid=[0, 1],
            command=self.toggle_lights,
            height=72,
            width=72,
        )

        self.siren_box = siren_box = Box(lower_box, layout="grid", border=2, align="left")
        _ = Text(siren_box, text="Siren", grid=[0, 0], size=16, underline=True)
        self.siren_button = PushButton(
            siren_box,
            image=self.siren_off,
            grid=[0, 1],
            height=72,
            width=72,
        )
        self.siren_button.when_clicked = lambda x: self.siren_req.send()
        self.siren_button.when_left_button_pressed = lambda _: self.toggle_sound(self.siren_button)
        self.siren_button.when_left_button_released = lambda _: self.toggle_sound(self.siren_button)

        self.klaxon_box = klaxon_box = Box(lower_box, layout="grid", border=2, align="left")
        _ = Text(klaxon_box, text="Klaxon", grid=[0, 0], size=16, underline=True)
        self.klaxon_button = PushButton(
            klaxon_box,
            image=self.siren_off,
            grid=[0, 1],
            height=72,
            width=72,
        )
        self.klaxon_button.when_clicked = self.toggle_klaxon

        self.gantry_box = gantry_box = Box(lower_box, layout="grid", border=2, align="left")
        _ = Text(gantry_box, text="Gantry", grid=[0, 0, 2, 1], size=16, underline=True)
        self.gantry_rev = PushButton(
            gantry_box,
            image=self.left_arrow,
            grid=[0, 1],
            height=70,
            width=70,
        )
        self.gantry_rev.when_clicked = lambda x: self.gantry_rev_req.send(repeat=2)

        self.gantry_fwd = PushButton(
            gantry_box,
            image=self.right_arrow,
            grid=[1, 1],
            height=70,
            width=70,
        )
        self.gantry_fwd.when_clicked = lambda x: self.gantry_fwd_req.send(repeat=2)

        # start upper box disabled
        self.upper_box.disable()
        self.lights_box.disable()
        self.siren_box.disable()
        self.klaxon_box.disable()
        self.gantry_box.disable()

        # sync GUI with current state
        self.sync_gui_state()

        # display GUI and start event loop; call blocks
        self.app.display()

    def reset(self):
        self.app.destroy()

    def update_counter(self, value: int = None):
        with self._cv:
            prefix = "-"
            if value is None:
                self.counter -= 1
            else:
                self.counter = value

            count = self.counter if self.counter is not None else 0
            if -30 <= count < 0:
                prefix = "+"
                count = abs(count)
                self.label.value = "Launch"
            else:
                self.label.value = "T-Minus"
                if count <= -30:
                    count = 0
                    if self._is_countdown is True:
                        self.count.cancel(self.update_counter)
                        self._is_countdown = False
                        self.launch.enable()
            minute = count // 60
            second = count % 60
            self.count.value = f"{prefix}{minute:02d}:{second:02d}"

    def do_launch(self, t_minus: int = 80, detected: bool = False, hold=False):
        with self._cv:
            print(f"Launching: T Minus: {t_minus}")
            if self._is_countdown is True:
                self.count.cancel(self.update_counter)
            self._is_countdown = True
            if detected is True:
                self.gantry_rev_req.send()
                self.siren_req.send()
            else:
                self.launch_seq_act()
            self.abort.enable()
            self.launch.disable()
            self.message.clear()
            self.update_counter(value=t_minus)
            # start the clock
            if hold is False:
                self.count.repeat(1090, self.update_counter)

    def do_abort(self, detected: bool = False):
        with self._cv:
            if detected is False:
                self.reset_req.send()
            self.message.clear()
            if self._is_countdown is True:
                self.count.cancel(self.update_counter)
                self._is_countdown = False
                if self.counter >= 0:
                    self.message.value = "Launch Aborted"
                else:
                    self.message.value = "Self Destruct"
            self.launch.enable()
            self.message.show()

    def toggle_power(self):
        self.update_counter(value=0)
        self.label.value = "T-Minus"
        self.message.clear()
        if self.power_button.image == self.on_button:
            self.do_power_on()
            self.power_on_req.send()
        else:
            self.do_power_off()
            self.power_off_req.send()
        self.power_button.height = self.power_button.width = 72

    def do_power_off(self):
        with self._cv:
            if self._is_countdown is True:
                self.count.cancel(self.update_counter)
                self._is_countdown = False
            if self.power_button.image != self.on_button:
                self.power_button.image = self.on_button
                self.power_button.height = self.power_button.width = 72
            self.upper_box.disable()
            self.lights_box.disable()
            self.siren_box.disable()
            self.klaxon_box.disable()
            self.gantry_box.disable()

    def do_power_on(self):
        with self._cv:
            if self.power_button.image != self.off_button:
                self.power_button.image = self.off_button
                self.power_button.height = self.power_button.width = 72
            self.upper_box.enable()
            self.lights_box.enable()
            self.siren_box.enable()
            self.klaxon_box.enable()
            self.gantry_box.enable()
            if self._is_countdown is True:
                self.launch.disable()

    def do_lights_on(self):
        if self.lights_button.image != self.off_button:
            self.lights_button.image = self.off_button
            self.lights_button.height = self.lights_button.width = 72

    def do_lights_off(self):
        if self.lights_button.image != self.on_button:
            self.lights_button.image = self.on_button
            self.lights_button.height = self.lights_button.width = 72

    def toggle_klaxon(self) -> None:
        self.klaxon_req.send()
        self.toggle_sound(self.klaxon_button)

    def do_klaxon_off(self):
        if self.klaxon_button.image != self.siren_off:
            self.klaxon_button.image = self.siren_off
            self.klaxon_button.height = self.klaxon_button.width = 72

    def toggle_lights(self):
        if self.lights_button.image == self.on_button:
            self.do_lights_on()
            self.lights_on_req.send(repeat=2)
        else:
            self.do_lights_off()
            self.lights_off_req.send(repeat=2)

    def toggle_sound(self, button: PushButton):
        if button.enabled is True:
            if button.image == self.siren_off:
                button.image = self.siren_on
            else:
                button.image = self.siren_off
            button.height = button.width = 72



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
import platform
import subprocess

from .singleton import singleton


@singleton
class HostInfo:
    def __init__(self) -> None:
        if hasattr(self, "_initialized") and self._initialized:
            return
        else:
            self._initialized = True

        self._system = platform.system()
        self._release = platform.release()
        self._version = platform.version()
        self._machine = platform.machine()
        self._node = platform.node()
        self._pi_model = None
        self._total_memory = None
        self._used_memory = None
        self._free_memory = None

        # do pi-specific stuff
        result = subprocess.run("cat /proc/device-tree/model".split(), capture_output=True, text=True)
        if result.returncode == 0:
            self._pi_model = result.stdout.strip().rstrip("\x00")

        result = subprocess.run("free -m".split(), capture_output=True, text=True)
        if result.returncode == 0:
            values = result.stdout.strip().rstrip("\x00").split()
            self._total_memory = int(values[7])
            self._used_memory = int(values[8])
            self._free_memory = int(values[9])

    def __repr__(self) -> str:
        pm = f"{self.pi_model} " if self.pi_model else ""
        return f"{pm}{self._system} {self._release} {self._machine} {self._node}"

    @property
    def is_linux(self) -> bool:
        return self._system == "Linux"

    @property
    def is_windows(self) -> bool:
        return self._system == "Windows"

    @property
    def is_macosx(self) -> bool:
        return self._system == "Darwin"

    @property
    def is_pi(self) -> bool:
        return self.pi_model is not None and self.pi_model.lower().startswith("raspberry pi")

    @property
    def pi_model(self) -> str:
        return self._pi_model

    @property
    def total_memory(self) -> int:
        return self._total_memory

    @property
    def free_memory(self) -> int:
        results = self._memory_usage()
        return results[2] if results else None

    @property
    def used_memory(self) -> int:
        results = self._memory_usage()
        return results[1] if results else None

    def _memory_usage(self) -> tuple | None:
        if self.is_pi:
            result = subprocess.run("free -m".split(), capture_output=True, text=True)
            if result.returncode == 0:
                v = result.stdout.strip().rstrip("\x00").split()
                return int(v[7]), int(v[8]), int(v[9]), int(v[10]), int(v[11]), int(v[12])
        return None



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

import os
import socket

#
# PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#
# SPDX-License-Identifier: LPGL
#
from queue import Queue
from typing import Any


class PollableQueue(Queue):
    def __init__(self, maxsize) -> None:
        super().__init__(maxsize)

        # Create a pair of connected sockets
        if os.name == "posix":
            self._put_socket, self._get_socket = socket.socketpair()
        else:
            # compatibility on non-POSIX systems
            server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            server.bind(("127.0.0.1", 0))
            server.listen(1)
            self._put_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._put_socket.connect(server.getsockname())
            self._get_socket, _ = server.accept()
            server.close()

    def fileno(self) -> int:
        return self._get_socket.fileno()

    def put(self, item: Any, block: bool = True, timeout: float = None) -> None:
        super().put(item, block=block, timeout=timeout)
        self._put_socket.send(b"x")

    def get(self, block: bool = True, timeout: float = None) -> Any:
        self._get_socket.recv(1)
        return super().get(block=block, timeout=timeout)



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

import os
import subprocess
import sys
import tempfile
import textwrap
from argparse import ArgumentParser
from typing import List, Set, Tuple

from .. import get_version, is_linux, is_package
from .pytrain import PROGRAM_NAME

SETTINGS = {
    "net_names": 0,
    "i2c": 0,
    "ssh": 0,
    "pi4video": 1,
    "boot_splash": 1,
    "rpi_connect": 1,
    "spi": 1,
    "serial_hw": 1,
    "serial_cons": 1,
    "onewire": 1,
    "rgpio": 1,
}

SERVICES = [
    "ModemManager",
    "bluealsa",
    "bluetooth",
    "colord",
    "cups",
    "cupsd",
    "dbus-org.bluez",
    "dbus-org.freedesktop.ModemManager1",
    "hciuart",
    "lightdm",
    "packagekit",
    "pipewire",
    "pulseaudio",
    "rpi-connect",
    "rpi-connect",
    "rpi-connect-wayvnc",
    "rpi-connect-wayvnc-watcher",
    "rpicam-apps",
]

PACKAGES = [
    "squeekboard",  # must be removed before labwc...
    "libgl1",
    "libegl1",
    "mesa-libgallium",
    "bluez",
    "colord",
    "cups",
    "cups-browsed",
    "dbus-org.freedesktop.ModemManager1",
    "firefox",
    "labwc",
    "mailcap",
    "mailutils",
    "modemmanager",
    "packagekit",
    "pipewire",
    "pulseaudio",
    "raindrop",
    "rpi-connect",
    "rpicam-apps",
]


class PiConfig:
    def __init__(self, cmd_line: List[str] = None) -> None:
        if cmd_line:
            args = self.command_line_parser().parse_args(cmd_line)
        else:
            args = self.command_line_parser().parse_args()
        self._args = args
        self.option = args.option
        self.verbose = args.quiet is False
        self.enable_spi = args.enable_spi
        if is_linux() is False:
            print("This command can only run on Raspberry Pi systems! Exiting...")
            sys.exit(1)
        # do the work
        if self.option == "check":
            self.do_check()
        elif self.option == "version":
            print(f"{self.__class__.__name__} {get_version()}")
        elif self.option == "expand_file_system":
            self.expand_file_system()
        else:
            do_reboot_msg = True
            if self.option == "all":
                cfg, svcs, pkgs = self.do_check("all")
            else:
                cfg = svcs = pkgs = None
            if self.option in {"all", "configuration"}:
                self.optimize_config(cfg)
                do_reboot_msg = False
            if self.option in {"all", "services"}:
                self.optimize_services(svcs)
                do_reboot_msg = False
            if self.option in {"all", "packages"}:
                self.optimize_packages(pkgs)
                do_reboot_msg = False
            if do_reboot_msg:
                print(f"Unknown optimization option: {self.option}")
            else:
                print("Your Pi must now be rebooted (sudo reboot)...")

    def do_check(self, option: str = "all") -> Tuple[Set[str], Set[str], Set[str]]:
        do_output = self.verbose is True and option == "all"
        cfg: Set[str] = set()
        if option in {"all", "configuration"}:
            if do_output:
                print("Checking Raspberry Pi Configuration...")
            for setting, value in SETTINGS.items():
                if setting == "spi" and self.enable_spi is True:
                    value = 0
                if do_output:
                    print(f"Checking {setting}...", end="")
                cmd = f"sudo raspi-config nonint get_{setting}"
                result = subprocess.run(cmd.split(), capture_output=True, text=True)
                if result.returncode == 0:
                    status = result.stdout.strip()
                    if status == str(value):
                        if do_output:
                            print("...OK")
                    else:
                        cfg.add(setting)
                        if do_output:
                            if self.verbose:
                                good = "ENABLED" if int(status) == 0 else "DISABLED"
                                print(f"...FAILED: {good}")
                            else:
                                good = "ENABLED" if value == 0 else "DISABLED"
                                bad = "DISABLED" if good == "ENABLED" else "ENABLED"
                                print(f"!!! {setting} is {bad} ({status}), should be {good} ({value})!!!")
                elif do_output:
                    if self.verbose:
                        print("...ERROR")
                    print(f"*** Check {setting} Error: {result.stderr.strip()} ***")
            if do_output:
                print("Checking /boot/firmware/config.txt...", end="")
            lines = self._read_config("/boot/firmware/config.txt")
            if lines:
                bluetooth_disabled = audio_disabled = camera_disabled = display_disabled = False
                for line in lines:
                    if line.startswith("dtoverlay=disable-bt"):
                        bluetooth_disabled = True
                    elif self._contains(line, "dtparam=audio=off", "#dtparam=audio=on", "# dtparam=audio=on"):
                        audio_disabled = True
                    elif self._contains(line, "dtparam=audio=on"):
                        audio_disabled = False
                    elif self._contains(line, "#camera_auto_detect=1", "# camera_auto_detect=1"):
                        camera_disabled = True
                    elif self._contains(line, "camera_auto_detect=1"):
                        camera_disabled = False
                    elif self._contains(line, "#display_auto_detect=1", "# display_auto_detect=1"):
                        display_disabled = True
                    elif self._contains(line, "display_auto_detect=1"):
                        display_disabled = False
                if bluetooth_disabled and audio_disabled and camera_disabled and display_disabled:
                    if do_output:
                        print("...OK")
                elif do_output:
                    cfg.add("config.txt")
                    if self.verbose:
                        print("...FAILED")
                    if bluetooth_disabled is False:
                        print("*** Bluetooth should be disabled ***")
                    if audio_disabled is False:
                        print("*** Audio should be disabled ***")
                    if camera_disabled is False:
                        print("*** Camera Autodetect should be disabled ***")
                    if display_disabled is False:
                        print("*** Display Autodetect should be disabled ***")
                else:
                    cfg.add("config.txt")
            else:
                if do_output:
                    if self.verbose:
                        print("...FAILED")
                    print("*** /boot/firmware/config.txt is empty ***")
        # check services
        svsc: Set[str] = set()
        if option in {"all", "services"}:
            if do_output:
                print("\nChecking installed services...")
            for service in SERVICES:
                if do_output:
                    print(f"Checking {service}...", end="")
                cmd = f"sudo systemctl status {service}.service"
                result = subprocess.run(cmd.split(), capture_output=True)
                if result.returncode == 4 or os.path.exists(f"/etc/systemd/system/{service}.service") is False:
                    if do_output:
                        print("...OK")
                else:
                    svsc.add(service)
                    if do_output:
                        if self.verbose:
                            print("...FOUND; can be removed")
                        else:
                            print(
                                f"*** {service} is installed; for {PROGRAM_NAME}, it can be deactivated and removed ***"
                            )
        # check packages
        pkgs: Set[str] = set()
        if option in {"all", "packages"}:
            if do_output:
                print("\nChecking packages...")
            for package in PACKAGES:
                if do_output:
                    print(f"Checking {package}...", end="")
                cmd = f"sudo apt policy {package}"
                result = subprocess.run(cmd.split(), capture_output=True, text=True)
                success = (
                    result.returncode != 0 or len(result.stdout.strip()) == 0 or "Installed: (none)" in result.stdout
                )
                if success:
                    if do_output:
                        print("...OK")
                else:
                    pkgs.add(package)
                    if do_output:
                        if self.verbose:
                            print("...CAN BE REMOVED")
                        else:
                            print(f"*** {package} installed; {PROGRAM_NAME} doesn't require it ***")
        return cfg, svsc, pkgs

    def optimize_config(self, cfg: Set[str] = None) -> None:
        for setting, value in SETTINGS.items():
            if setting == "spi" and self.enable_spi is True:
                value = 0
            cmd = f"sudo raspi-config nonint do_{setting} {value}"
            if self.verbose:
                print(f"Executing: {cmd}...", end="")
            try:
                result = subprocess.run(cmd.split(), capture_output=True)
                if result.returncode == 0:
                    if self.verbose:
                        print("...Done")
                else:
                    print(f"...Failed with code {result.returncode}: {result.stderr.decode('utf-8').strip()}")
            except Exception as e:
                print(e)
        if (cfg and "config.txt" in cfg) or cfg is None:
            reboot_required = False
            print("Updating /boot/firmware/config.txt...")
            lines = self._read_config("/boot/firmware/config.txt", make_copy=True)
            if lines:
                bluetooth_disabled = False
                newlines = {}
                for i, line in enumerate(lines):
                    if line.startswith("dtoverlay=disable-bt"):
                        bluetooth_disabled = True
                    elif self._contains(line, "dtparam=audio=on"):
                        newlines[i] = "dtparam=audio=off"
                    elif self._contains(line, "camera_auto_detect=1"):
                        newlines[i] = "#camera_auto_detect=1"
                    elif self._contains(line, "display_auto_detect=1"):
                        newlines[i] = "#display_auto_detect=1"
                # replace modified lines
                for k, v in newlines.items():
                    reboot_required = True
                    lines[k] = v

                # disable bluetooth, if needed
                if bluetooth_disabled is False:
                    reboot_required = True
                    lines.append("")
                    lines.append("# Disable Bluetooth")
                    lines.append("dtoverlay=disable-bt")

                if reboot_required:
                    # rewrite the file
                    tmp = tempfile.NamedTemporaryFile()
                    with open(tmp.name, "w") as f:
                        for line in lines:
                            f.write(f"{line}\n")
                    subprocess.run(f"sudo cp -f {tmp.name} /boot/firmware/config.txt".split())
                    print("*** Reboot required to apply changes...")

    def optimize_services(self, svsc: Set[str] = None) -> None:
        if svsc is None or len(svsc) > 0:
            if self.verbose:
                print("Disabling/removing unneeded services...")
            for service in SERVICES:
                results = list()
                if self.verbose:
                    print(f"Disabling: {service} service...", end="")
                for sub_cmd in ["stop", "disable"]:
                    cmd = f"sudo systemctl {sub_cmd} {service}.service"
                    try:
                        results.append(subprocess.run(cmd.split(), capture_output=True))
                    except Exception as e:
                        print(f"Error disabling {service}: {e}")
                # delete service file, if it exists
                if os.path.exists(f"/etc/systemd/system/{service}.service"):
                    subprocess.run(f"sudo rm -f /etc/systemd/system/{service}.service".split())
                if self.verbose:
                    print("...Done")
            # do a daemon reload
            if self.verbose:
                print("Reloading daemon services...")
            subprocess.run("sudo systemctl daemon-reload".split())
        else:
            if self.verbose:
                print("No unneeded services found!")

    def optimize_packages(self, pkgs: Set[str] = None):
        if pkgs is None:
            _, _, pkgs = self.do_check("packages")
        if len(pkgs) == 0:
            if self.verbose:
                print("No extraneous packages remain!")
            return

        # hack to deal with squeekboard/labwc issue
        if os.path.isfile("/usr/share/labwc/autostart"):
            pass
        else:
            subprocess.run("sudo mkdir -p /usr/share/labwc".split())
            subprocess.run("sudo touch /usr/share/labwc/autostart".split())

        if self.verbose:
            text = ", ".join(pkgs)
            text = (
                f"The following packages aren't needed to run {PROGRAM_NAME} and will be removed: {text}; "
                f"This may take awhile..."
            )
            print(textwrap.fill(text, width=70))

        cmd = f"sudo apt purge -y {' '.join(pkgs)}"
        try:
            subprocess.run(cmd.split())
        except Exception as e:
            print(f"Error removing packages: {e}")

        if self.verbose:
            print("Removing unused files... This may take a while...")
        r = subprocess.run("sudo apt autoremove --purge -y".split(), capture_output=True, text=True)
        if self.verbose:
            print(r.stdout.strip())
        # cups (printing damon) is particularly pernicious...
        cmd = "sudo find . -name 'cups*' -print | sudo xargs rm -fr"
        r = subprocess.run(cmd.split(), capture_output=True, text=True)
        if self.verbose:
            print(r.stdout.strip())

    @staticmethod
    def expand_file_system():
        print("Expanding file system; your Pi will reboot...")
        cmd = "sudo sudo raspi-config --expand-rootfs"
        result = subprocess.run(cmd.split(), capture_output=True, text=True)
        if result.returncode != 0:
            print(f"Error expanding file system: {result.stderr.strip()}")
        else:
            print("File system expansion complete! Rebooting...")
            subprocess.run("sudo reboot".split(), capture_output=True, text=True)

    def _read_config(self, filename: str = "/boot/firmware/config.txt", make_copy: bool = False) -> List[str]:
        config = list()
        if os.path.exists(filename):
            if make_copy:
                if self.verbose:
                    print(f"Making a backup copy of {filename}: {filename}.bak...")
                subprocess.run(f"sudo cp -f {filename} {filename}.bak".split())
            with open(filename, "r") as f:
                for line in f:
                    config.append(line.strip())
        return config

    def command_line_parser(self) -> ArgumentParser:
        prog = "piconfig" if is_package() else "piconfig.py"
        parser = ArgumentParser(
            prog=prog,
            description=f"Optimize Raspberry Pi for use with {PROGRAM_NAME}",
        )
        config_group = parser.add_mutually_exclusive_group()
        config_group.add_argument(
            "-check",
            action="store_const",
            const="check",
            dest="option",
            help="Check Raspberry Pi configuration (no changes made; default option)",
        )
        config_group.add_argument(
            "-all",
            action="store_const",
            const="all",
            dest="option",
            help="Perform all optimizations",
        )
        config_group.add_argument(
            "-configuration",
            action="store_const",
            const="configuration",
            dest="option",
            help="Optimize Raspberry Pi configuration options",
        )

        config_group.add_argument(
            "-packages",
            action="store_const",
            const="packages",
            dest="option",
            help="Optimize packages",
        )
        config_group.add_argument(
            "-services",
            action="store_const",
            const="services",
            dest="option",
            help="Optimize services",
        )
        parser.set_defaults(option="check")

        misc_opts = parser.add_argument_group("Miscellaneous options")
        misc_opts.add_argument(
            "-expand_file_system",
            action="store_const",
            const="expand_file_system",
            dest="option",
            help="Expand file system and reboot",
        )
        misc_opts.add_argument(
            "-enable_spi",
            action="store_true",
            help="Enable the SPI interface (and don't try to disable it)",
        )
        misc_opts.add_argument(
            "-quiet",
            action="store_true",
            help="Operate quietly and don't provide feedback",
        )
        misc_opts.add_argument(
            "-version",
            action="version",
            version=f"{self.__class__.__name__} {get_version()}",
            help="Show version and exit",
        )
        return parser

    @staticmethod
    def _contains(line: str, *args: str) -> bool:
        for p in args:
            if line.startswith(p):
                return True
        return False


def main(args: list[str] | None = None) -> int:
    if args is None:
        args = sys.argv[1:]
    try:
        PiConfig(args)
        return 0
    except Exception as e:
        # Output anything else nicely formatted on stderr and exit code 1
        sys.exit(f"{__file__}: error: {e}\n")



from threading import RLock
from typing import Tuple

from gpiozero import GPIODeviceClosed, HoldMixin

from .i2c_device import I2CDevice
from .mcp23017 import INPUT, HIGH, Mcp23017Factory


class ButtonI2C(I2CDevice, HoldMixin):
    def __init__(
        self,
        pin: int | Tuple[int, int] | Tuple[int, int, int],
        i2c_address: int = 0x23,
        pull_up: bool = True,
        bounce_time: float = None,
        hold_time: float = 1,
        hold_repeat: bool = False,
        interrupt_pin: int | str = None,
        pin_factory=None,
    ):
        self._lock = RLock()
        # i2c buttons use the MCP 23017 i2c dio board, which supports 16 pins and interrupts
        if isinstance(pin, tuple):
            if len(pin) > 1 and pin[1]:
                interrupt_pin = pin[1]
            if len(pin) > 2 and pin[2]:
                i2c_address = pin[2]
            if len(pin) > 3 and pin[3] in {True, False}:
                pull_up = pin[3]
            pin = pin[0]
        self._dio_pin = pin
        self._mcp_23017 = Mcp23017Factory.build(
            address=i2c_address,
            pin=pin,
            interrupt_pin=interrupt_pin,
            client=self,
        )
        # initialize the gpiozero device
        super().__init__(pin_factory=pin_factory)

        # configure the Mcp23017 pin to the appropriate mode
        self._mcp_23017.set_pin_mode(pin, INPUT)
        self._mcp_23017.set_pull_up(pin, pull_up)
        if interrupt_pin is not None:
            self._mcp_23017.enable_interrupt(pin)
        self._interrupt_pin = interrupt_pin
        self._bounce_time = bounce_time
        self.hold_time = hold_time if hold_time is not None and hold_time >= 0 else 1
        self.hold_repeat = hold_repeat

        # Call _fire_events once to set initial state of events
        self._fire_events(self.pin_factory.ticks(), self.is_active)

    def __repr__(self):
        # noinspection PyBroadException
        try:
            return (
                f"<{self.__class__.__name__} object on pin "
                f"{self.pin!r}, pull_up={self.pull_up}, "
                f"is_active={self.is_active}>"
            )
        except Exception:
            return super().__repr__()

    @property
    def pin(self) -> int:
        if self._mcp_23017 is None:
            raise GPIODeviceClosed("I2C Button is closed or uninitialized")
        return self._dio_pin

    def close(self) -> None:
        with self._lock:
            try:
                # in edge cases where constructor fails, _mcp_23017 property may not exist
                if hasattr(self, "_mcp_23017") and self._mcp_23017 is not None:
                    self._mcp_23017.disable_interrupt(self._dio_pin)
                    self._mcp_23017.deregister_client(self)
                    Mcp23017Factory.close(self._mcp_23017, self._dio_pin)
                self._mcp_23017 = None
            finally:
                super().close()

    @property
    def closed(self) -> bool:
        return self._mcp_23017 is None

    @property
    def bounce_time(self) -> float:
        return self._bounce_time

    @bounce_time.setter
    def bounce_time(self, value: float) -> None:
        if value is not None and value < 0:
            raise ValueError("bounce_time must be None or >= 0")
        self._bounce_time = value

    def _signal_event(self, active: bool) -> None:
        self._fire_events(self.pin_factory.ticks(), active)

    @property
    def i2c_address(self) -> int:
        return self._mcp_23017.address

    @property
    def value(self) -> int:
        if self._mcp_23017 is None:
            raise GPIODeviceClosed("I2C Button is closed or uninitialized")
        if self.pull_up:
            return 0 if self._mcp_23017.digital_read(self._dio_pin) == HIGH else 1
        else:
            return 1 if self._mcp_23017.digital_read(self._dio_pin) == HIGH else 0

    @property
    def is_active(self) -> bool:
        if self._mcp_23017 is None:
            raise GPIODeviceClosed("I2C Button is closed or uninitialized")
        return self.value == 1

    @property
    def pull_up(self) -> bool:
        if self._mcp_23017 is None:
            raise GPIODeviceClosed("I2C Button is closed or uninitialized")
        return self._mcp_23017.get_pull_up(self._dio_pin)


ButtonI2C.is_pressed = ButtonI2C.is_active
ButtonI2C.pressed_time = ButtonI2C.active_time
ButtonI2C.when_pressed = ButtonI2C.when_activated
ButtonI2C.when_released = ButtonI2C.when_deactivated
ButtonI2C.wait_for_press = ButtonI2C.wait_for_active
ButtonI2C.wait_for_release = ButtonI2C.wait_for_inactive



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
from ..protocol.command_req import CommandReq
from ..protocol.constants import CommandScope
from ..protocol.tmcc1.tmcc1_constants import TMCC1AuxCommandEnum
from .gpio_device import GpioDevice, P
from .state_source import AccessoryStateSource


class GantryCrane(GpioDevice):
    """
    Represent a Gantry Crane device with multiple configurable controls and states.

    This class extends the functionality of a GPIO-based device to provide a set of commands
    and configurations for controlling various parts of a gantry crane. These include movement
    of the crane's cab, boom, roll, and magnetic operations. The class allows interaction through
    different input pins and can integrate additional features like rotary encoders, LEDs, and
    accessory states for comprehensive control.

    Attributes:
        cab_left_btn (Union[Button, None]): Represents the button control for rotating the cab left.
        cab_right_btn (Union[Button, None]): Represents the button control for rotating the cab right.
        cab_re (Union[PyRotaryEncoder, None]): Represents the rotary encoder for cab rotation
            if enabled, otherwise None.
        ro_left_btn (Union[Button, None]): Represents the button control for rolling left,
            if configured.
        ro_right_btn (Union[Button, None]): Represents the button control for rolling right,
            if configured.
        down_btn (Union[Button, None]): Represents the button control for moving the boom
            downward, if configured.
        up_btn (Union[Button, None]): Represents the button control for lifting the boom,
            if configured.
        mag_btn (Union[Button, None]): Represents the toggle button for magnetic operation,
            if configured.
        mag_led (Union[LED, None]): Represents the LED linked to magnetic operation,
            if configured.

    Args:
        address (int): The unique address of the gantry crane device.
        cab_left_pin (P): GPIO pin assigned for the cab rotation left control.
        cab_right_pin (P): GPIO pin assigned for the cab rotation right control.
        ro_left_pin (P): Optional; GPIO pin assigned for rolling left control. Default is None.
        ro_right_pin (P): Optional; GPIO pin assigned for rolling right control. Default is None.
        bo_down_pin (P): Optional; GPIO pin assigned for moving the boom downward. Default is None.
        bo_up_pin (P): Optional; GPIO pin assigned for lifting the boom. Default is None.
        mag_pin (P): Optional; GPIO pin assigned for the magnetic operation control. Default is None.
        led_pin (P): Optional; GPIO pin assigned for the LED connected to the magnetic operation.
            Default is None.
        cathode (bool): Optional; Specifies whether the LED uses common cathode configuration.
            Default is True.
        cab_rotary_encoder (bool): Optional; Determines if a rotary encoder is used for
            cab rotation. Default is False.

    Raises:
        No raised errors are explicitly documented.
    """

    def __init__(
        self,
        address: int,
        cab_left_pin: P,
        cab_right_pin: P,
        ro_left_pin: P = None,
        ro_right_pin: P = None,
        bo_down_pin: P = None,
        bo_up_pin: P = None,
        mag_pin: P = None,
        led_pin: P = None,
        cathode: bool = True,
        cab_rotary_encoder: bool = False,
        repeat_every: float = 0.02,
    ) -> None:
        cab_sel_cmd = CommandReq.build(TMCC1AuxCommandEnum.NUMERIC, address, data=1, scope=CommandScope.ACC)
        if cab_rotary_encoder:
            from .py_rotary_encoder import PyRotaryEncoder

            self.cab_left_btn = self.cab_right_btn = None
            cmd = CommandReq.build(TMCC1AuxCommandEnum.RELATIVE_SPEED, address, data=0, scope=CommandScope.ACC)
            self.cab_re = PyRotaryEncoder(
                cab_left_pin,
                cab_right_pin,
                cmd,
                wrap=False,
                initial_step=0,
                max_steps=180,
                steps_to_data=self.std_step_to_data,
                pause_for=repeat_every,
                reset_after_motion=True,
                prefix_cmd=cab_sel_cmd,
            )
        else:
            # use momentary contact switch to rotate cab
            self.cab_re = None
            left_cmd, self.cab_left_btn, _ = self.make_button(
                cab_left_pin,
                command=TMCC1AuxCommandEnum.RELATIVE_SPEED,
                address=address,
                data=-1,
                scope=CommandScope.ACC,
                hold_repeat=True,
                hold_time=repeat_every,
            )
            self.cab_left_btn.when_pressed = self.with_prefix_action(cab_sel_cmd, left_cmd)
            self.cab_left_btn.when_held = left_cmd.as_action()

            right_cmd, self.cab_right_btn, _ = self.make_button(
                cab_right_pin,
                command=TMCC1AuxCommandEnum.RELATIVE_SPEED,
                address=address,
                data=1,
                scope=CommandScope.ACC,
                hold_repeat=True,
                hold_time=repeat_every,
            )
            self.cab_right_btn.when_pressed = self.with_prefix_action(cab_sel_cmd, right_cmd)
            self.cab_right_btn.when_held = right_cmd.as_action()

        # set up commands for roll
        ro_sel_cmd = CommandReq.build(TMCC1AuxCommandEnum.NUMERIC, address, data=2, scope=CommandScope.ACC)
        # roll left
        if ro_left_pin:
            ro_left_cmd, self.ro_left_btn, _ = self.make_button(
                ro_left_pin,
                command=TMCC1AuxCommandEnum.RELATIVE_SPEED,
                address=address,
                data=-1,
                scope=CommandScope.ACC,
                hold_repeat=True,
                hold_time=repeat_every,
            )
            self.ro_left_btn.when_pressed = self.with_prefix_action(ro_sel_cmd, ro_left_cmd)
            self.ro_left_btn.when_held = ro_left_cmd.as_action()
        else:
            self.ro_left_btn = None

        # roll right
        if ro_right_pin:
            ro_right_cmd, self.ro_right_btn, _ = self.make_button(
                ro_right_pin,
                command=TMCC1AuxCommandEnum.RELATIVE_SPEED,
                address=address,
                data=1,
                scope=CommandScope.ACC,
                hold_repeat=True,
                hold_time=repeat_every,
            )
            self.ro_right_btn.when_pressed = self.with_prefix_action(ro_sel_cmd, ro_right_cmd)
            self.ro_right_btn.when_held = ro_right_cmd.as_action()
        else:
            self.ro_right_btn = None

        # set up commands for boom down
        if bo_down_pin:
            down_cmd, self.down_btn, _ = self.make_button(
                bo_down_pin,
                TMCC1AuxCommandEnum.BRAKE_SPEED,
                address,
                hold_repeat=True,
                hold_time=repeat_every,
            )
            self.down_btn.when_pressed = down_cmd.as_action()
            self.down_btn.when_held = down_cmd.as_action()
        else:
            self.down_btn = None

        # boom lift
        if bo_up_pin:
            up_cmd, self.up_btn, _ = self.make_button(
                bo_up_pin,
                TMCC1AuxCommandEnum.BOOST_SPEED,
                address,
                hold_repeat=True,
                hold_time=repeat_every,
            )
            self.up_btn.when_pressed = up_cmd.as_action()
            self.up_btn.when_held = up_cmd.as_action()
        else:
            self.up_btn = None

        if mag_pin is not None:
            self.mag_btn, self.mag_led = self.when_toggle_button_pressed(
                mag_pin,
                TMCC1AuxCommandEnum.AUX2_OPTION_ONE,
                address,
                led_pin=led_pin,
                auto_timeout=59,
                cathode=cathode,
            )
            self.mag_led.blink()
            self.cache_handler(
                AccessoryStateSource(
                    address,
                    self.mag_led,
                    aux2_state=TMCC1AuxCommandEnum.AUX2_OPTION_ONE,
                )
            )
        else:
            self.mag_btn = self.mag_led = None



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
from abc import ABC, ABCMeta, abstractmethod
from threading import Thread
from typing import Callable, TypeVar, Union

from gpiozero import LED, Button, Device

from ..protocol.command_def import CommandDefEnum
from ..protocol.command_req import CommandReq
from ..protocol.constants import DEFAULT_ADDRESS, CommandScope
from .gpio_handler import GpioHandler
from .py_rotary_encoder import PyRotaryEncoder

P = TypeVar("P", bound=Union[int, str, tuple[int], tuple[int, int], tuple[int, int, int]])


class GpioDevice(ABC):
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self):
        pass

    def close(self):
        for k, v in self.__dict__.items():
            if isinstance(v, Device):
                v.close()
            if isinstance(v, PyRotaryEncoder):
                v.close()
                v.reset()

    @staticmethod
    def ramped_speed(step_no: int, max_steps: int = 180) -> int:
        speed = 0
        if step_no < 0:
            mag = -1
            step_no = abs(step_no)
        else:
            mag = 1
        if step_no > 0:
            third = max_steps / 3
            if step_no <= third:
                speed = 1
            elif step_no <= 2 * third:
                speed = 2
            else:
                speed = 3
        print(step_no * mag)
        return speed * mag

    @staticmethod
    def std_step_to_data(step_no: int) -> int:
        return 1 if step_no > 0 else -1 if step_no < 0 else 0

    @staticmethod
    def fast_step_to_data(step_no: int) -> int:
        return 3 if step_no > 0 else -3 if step_no < 0 else 0

    @staticmethod
    def make_button(
        pin: P,
        command: CommandReq | CommandDefEnum = None,
        address: int = DEFAULT_ADDRESS,
        data: int = None,
        scope: CommandScope = None,
        led_pin: P = None,
        hold_repeat: bool = False,
        hold_time: float = 1,
        initially_on: bool = False,
        bind: bool = False,
        cathode: bool = True,
    ) -> Button | tuple[CommandReq, Button, LED]:
        return GpioHandler.make_button(
            pin=pin,
            command=command,
            address=address,
            data=data,
            scope=scope,
            led_pin=led_pin,
            hold_repeat=hold_repeat,
            hold_time=hold_time,
            initially_on=initially_on,
            bind=bind,
            cathode=cathode,
        )

    @staticmethod
    def make_led(
        pin: P,
        initially_on: bool = False,
        cathode: bool = True,
    ) -> LED:
        return GpioHandler.make_led(
            pin=pin,
            initially_on=initially_on,
            cathode=cathode,
        )

    @staticmethod
    def cache_handler(handler: Thread) -> None:
        GpioHandler.GPIO_HANDLER_CACHE.add(handler)

    @staticmethod
    def cache_device(device: Device) -> None:
        """
        Keep devices around after creation so they remain in scope
        """
        GpioHandler.GPIO_DEVICE_CACHE.add(device)

    @staticmethod
    def when_button_pressed(
        pin: P,
        command: CommandReq | CommandDefEnum,
        address: int = DEFAULT_ADDRESS,
        data: int = 0,
        scope: CommandScope = None,
        led_pin: P = None,
        cathode: bool = True,
    ) -> tuple[CommandReq, Button, LED]:
        return GpioHandler.when_button_pressed(
            pin=pin,
            command=command,
            address=address,
            data=data,
            scope=scope,
            led_pin=led_pin,
            cathode=cathode,
        )

    @staticmethod
    def when_toggle_button_pressed(
        pin: P,
        command: CommandReq | CommandDefEnum,
        address: int = DEFAULT_ADDRESS,
        data: int = 0,
        scope: CommandScope = None,
        led_pin: P = None,
        initial_state: bool = False,
        auto_timeout: int = None,
        cathode: bool = True,
    ) -> tuple[Button, LED]:
        return GpioHandler.when_toggle_button_pressed(
            pin=pin,
            command=command,
            address=address,
            data=data,
            scope=scope,
            led_pin=led_pin,
            initial_state=initial_state,
            auto_timeout=auto_timeout,
            cathode=cathode,
        )

    @staticmethod
    def with_on_action(action: Callable, led: LED, *impacted_leds: LED) -> Callable:
        return GpioHandler.with_on_action(action, led, *impacted_leds)

    @staticmethod
    def with_off_action(action: Callable, led: LED, *impacted_leds: LED) -> Callable:
        return GpioHandler.with_off_action(action, led, *impacted_leds)

    @staticmethod
    def with_prefix_action(
        prefix: CommandReq,
        command: CommandReq,
    ) -> Callable:
        return GpioHandler.with_prefix_action(prefix, command)



from __future__ import annotations

from .sequence_req import SequenceReq
from ..command_req import CommandReq
from ..constants import CommandScope
from ..sequence.sequence_constants import SequenceCommandEnum
from ..tmcc2.tmcc2_constants import TMCC2EngineCommandEnum


class GradeCrossingReq(SequenceReq):
    def __init__(self, address: int, data: int, scope: CommandScope = CommandScope.ENGINE) -> None:
        super().__init__(SequenceCommandEnum.GRADE_CROSSING_SEQ, address, scope)
        self._data = data
        req15 = CommandReq(TMCC2EngineCommandEnum.QUILLING_HORN, address, 15, scope)
        req8 = CommandReq(TMCC2EngineCommandEnum.QUILLING_HORN, address, 8, scope)
        req4 = CommandReq(TMCC2EngineCommandEnum.QUILLING_HORN, address, 4, scope)
        req0 = CommandReq(TMCC2EngineCommandEnum.QUILLING_HORN, address, 0, scope)
        # first blast
        for _ in range(5):
            self.add(req15)
        self.add(req0)
        self.add(req8)

        # second blast
        delay = 1.1
        for _ in range(4):
            self.add(req8, delay=delay)
        self.add(req0, delay=delay)
        for _ in range(2):
            self.add(req8, delay=delay)

        # third blast
        delay += 1.1
        self.add(req8, delay=delay)
        self.add(req0, delay=delay)
        for _ in range(6):
            self.add(req15, delay=delay)

        # fourth blast
        delay += 1.05
        for _ in range(6):
            self.add(req15, delay=delay)
        for _ in range(3):
            self.add(req4, delay=delay)
        self.add(req0, delay=delay)



class MBusMock:
    def __init__(self) -> None:
        self.data = {}
        for address in range(0x20, 0x28):
            self.data[address] = {}
            for offset in range(0x00, 0x16):
                self.data[address][offset] = 0x00
                if offset == 0x00 or offset == 0x01:
                    self.data[address][offset] = 0xFF

    def write_byte_data(self, address, offset, value) -> None:
        self.data[address][offset] = value

    def read_byte_data(self, address, offset) -> int:
        return self.data[address][offset]

    def read_byte(self, address) -> int:
        return next(iter(self.data[address].values()))



from __future__ import annotations

import time
from collections import deque
from threading import Condition, Event
from typing import List, Callable

from gpiozero import Button, CompositeDevice, GPIOPinMissing, DigitalOutputDevice, event, EventsMixin
from gpiozero.threads import GPIOThread
from smbus2 import SMBus

from .gpio_handler import GpioHandler

DEFAULT_4X4_KEYS = (
    ("1", "2", "3", "A"),
    ("4", "5", "6", "B"),
    ("7", "8", "9", "C"),
    ("*", "0", "#", "D"),
)

TELEPHONE_4X3_KEYS = (
    ("1", "2", "3"),
    ("4", "5", "6"),
    ("7", "8", "9"),
    ("*", "0", "#"),
)


class Keypad(EventsMixin, CompositeDevice):
    def __init__(
        self,
        row_pins: List[int | str],
        column_pins: List[int | str],
        bounce_time: float = None,
        keys: List[List[str]] = DEFAULT_4X4_KEYS,
        clear_key: str = "C",
        digit_key: str = "D",
        eol_key: str = "#",
        pin_factory=None,
        key_queue: KeyQueue = None,
    ):
        if keys is None or len(keys) == 0:
            raise ValueError("Must specify at least one row of keys")
        if len(row_pins) != len(keys):
            raise GPIOPinMissing(f"Number of row pins must match the number of rows ({len(row_pins)} != {len(keys)})")
        num_cols = len(keys[0])
        if len(column_pins) != num_cols:
            raise ValueError(
                f"Number of column pins must match the number of keys ({len(column_pins)} != {len(keys[0])})"
            )

        devices = []
        self._rows = []
        for pin in row_pins:
            dev = DigitalOutputDevice(pin, pin_factory=pin_factory)
            self._rows.append(dev)
            devices.append(dev)

        self._cols = []
        for pin in column_pins:
            dev = Button(
                pin,
                pull_up=False,
                bounce_time=bounce_time,
                hold_repeat=False,
                active_state=None,
                pin_factory=pin_factory,
            )
            self._cols.append(dev)
            devices.append(dev)
        super().__init__(*devices, pin_factory=pin_factory)

        if len(self) == 0:
            raise GPIOPinMissing("No pins given")

        # _handlers only exist to ensure that we keep a reference to the
        # generated fire_both_events handler for each Button (remember that
        # pin.when_changed only keeps a weak reference to handlers)
        def get_new_handler(device):
            def fire_both_events(ticks, state):
                # noinspection PyProtectedMember
                device._fire_events(ticks, device._state_to_value(state))
                self._fire_events(ticks, self.is_active)

            return fire_both_events

        self._handlers = tuple(get_new_handler(device) for device in self._cols)
        for button, handler in zip(self._cols, self._handlers):
            button.pin.when_changed = handler

        self._when_changed = None
        self._last_value = None
        self._keypress = self._last_keypress = None
        self._keys = keys
        if key_queue is None:
            self._key_queue = KeyQueue(clear_key=clear_key, digit_key=digit_key, eol_key=eol_key)
        elif isinstance(key_queue, KeyQueue):
            self._key_queue = key_queue
        else:
            raise ValueError(f"{key_queue} is not a KeyQueue")
        self.when_pressed = self._key_queue.keypress_handler()

        # Call _fire_events once to set initial state of events
        self._fire_events(self.pin_factory.ticks(), self.is_active)

        # create the background thread to continually scan the matrix
        self._scan_thread = GPIOThread(self._scan_keyboard)
        self._scan_thread.daemon = True
        self._is_running = True
        self._scan_thread.start()
        GpioHandler.cache_handler(self._scan_thread)

    def reset(self) -> None:
        self.close()

    def close(self) -> None:
        self._is_running = False
        self._reset_pin_states()
        super().close()
        if self._key_queue:
            self._key_queue.reset()
        if self._scan_thread:
            self._scan_thread.stop()

    @property
    def keypress(self) -> str | None:
        return self._keypress

    @property
    def last_keypress(self) -> str | None:
        return self._last_keypress

    @property
    def key_queue(self) -> KeyQueue:
        return self._key_queue

    """
    The following methods expose behavior of KeyQueue and are
    repeated here for convenience
    """

    def reset_key_presses(self) -> None:
        self._key_queue.reset()
        self._keypress = self._last_keypress = None

    def wait_for_eol(self, timeout: float = 10) -> str | None:
        return self._key_queue.wait_for_eol(timeout)

    @property
    def key_presses(self) -> str:
        return self._key_queue.key_presses

    @property
    def is_eol(self) -> bool:
        return self._key_queue.is_eol

    when_changed = event()

    def _fire_changed(self):
        if self.when_changed:
            self.when_changed()

    def _fire_events(self, ticks, new_value):
        super()._fire_events(ticks, new_value)
        old_value, self._last_value = self._last_value, new_value
        if old_value is None:
            # Initial "indeterminate" value; don't do anything
            pass
        elif old_value != new_value:
            self._fire_changed()

    def _reset_pin_states(self) -> None:
        """
        Resets the states of all pins associated with the rows.

        Sets the state of each row pin to 'off' if it is not already closed.
        """
        for r in self._rows:
            if r.closed is False:
                r.off()

    def _scan_keyboard(self) -> None:
        """
        Scan the keys and handle keypress events in a loop that runs in a
        background thread.

        This method continuously scans the keys of a matrix keypad until
        the _is_running flag is set to False. For each row, it checks each
        column to detect if a key is pressed and handles keypress events
        accordingly.

        The method ensures a rest period to avoid excessive CPU usage
        when no key is pressed.
        """
        while self._is_running:
            self._reset_pin_states()
            self._keypress = None
            for r, row in enumerate(self._rows):
                row.on()
                try:
                    for c, col in enumerate(self._cols):
                        if col.is_active:
                            self._keypress = self._last_keypress = self._keys[r][c]
                            self._fire_events(self.pin_factory.ticks(), True)
                            while col.is_active:
                                time.sleep(0.05)
                            self._fire_events(self.pin_factory.ticks(), False)
                            break
                finally:
                    if self._keypress:
                        break
                    row.off()
            if self._keypress is None:
                time.sleep(0.05)  # give CPU a break


Keypad.is_pressed = Keypad.is_active
Keypad.pressed_time = Keypad.active_time
Keypad.when_pressed = Keypad.when_activated
Keypad.when_released = Keypad.when_deactivated
Keypad.wait_for_press = Keypad.wait_for_active
Keypad.wait_for_release = Keypad.wait_for_inactive


# PCF8574T address
KEYPAD_PCF8574_ADDRESS = 0x20

# Row pins on the PCF8574T
ROW_PINS = (0, 1, 2, 3)

# Column pins on the PCF8574T
COL_PINS = (4, 5, 6, 7)


class KeyPadI2C:
    def __init__(
        self,
        i2c_address: int = KEYPAD_PCF8574_ADDRESS,
        row_pins: List[int] = ROW_PINS,
        col_pins: List[int] = COL_PINS,
        keys: List[List[str]] = DEFAULT_4X4_KEYS,
        clear_key: str = "C",
        digit_key: str = "D",
        eol_key: str = "#",
        swap_key: str = "*",
        key_queue: KeyQueue = None,
    ):
        self._i2c_address = i2c_address
        self._row_pins = row_pins
        self._col_pins = col_pins
        self._is_running = True
        self._last_value = None
        self._keypress = self._last_keypress = None
        self._keys = keys
        if key_queue is None:
            self._key_queue = KeyQueue(
                clear_key=clear_key,
                digit_key=digit_key,
                eol_key=eol_key,
                swap_key=swap_key,
            )
        elif isinstance(key_queue, KeyQueue):
            self._key_queue = key_queue
        else:
            raise ValueError(f"{key_queue} is not a KeyQueue")
        self._keypress_handler = self._key_queue.keypress_handler()

        # create the background thread to continually scan the matrix
        self._scan_thread = GPIOThread(self._scan_keyboard)
        self._scan_thread.daemon = True
        self._is_running = True
        self._scan_thread.start()
        GpioHandler.cache_handler(self._scan_thread)

    def reset(self) -> None:
        self.close()

    def close(self) -> None:
        self._is_running = False
        if self._key_queue:
            self._key_queue.reset()
        if self._scan_thread:
            self._scan_thread.stop()

    @property
    def keypress(self) -> str | None:
        return self._keypress

    @property
    def last_keypress(self) -> str | None:
        return self._last_keypress

    @property
    def key_queue(self) -> KeyQueue:
        return self._key_queue

    def _scan_keyboard(self) -> None:
        """
        Scan the keys and handle keypress events in a loop that runs in a
        background thread.

        This method continuously scans the keys of a matrix keypad until
        the _is_running flag is set to False. For each row, it checks each
        column to detect if a key is pressed and handles keypress events
        accordingly.

        The method ensures a rest period to avoid excessive CPU usage
        when no key is pressed.
        """
        with SMBus(1) as bus:  # Use the appropriate I2C bus number
            while self._is_running is True and self._scan_thread.stopping.is_set() is False:
                key = self.read_keypad(bus)
                if key is not None:
                    self._last_keypress = self._keypress
                    self._keypress = key
                    self._keypress_handler(self)

    def read_keypad(self, bus: SMBus = None):
        if bus is None:
            bus = SMBus(1)
        """
        Reads the state of the matrix keypad. If a key is being pressed, return it
        once released. If no key is being pressed, return None.
        """
        for r, row_pin in enumerate(self._row_pins):
            bus.write_byte(self._i2c_address, 0xFF & ~(1 << row_pin))
            time.sleep(0.001)
            for c, col_pin in enumerate(self._col_pins):
                if bus.read_byte(self._i2c_address) & (1 << col_pin) == 0:
                    while bus.read_byte(self._i2c_address) & (1 << col_pin) == 0:
                        time.sleep(0.05)
                    return self._keys[r][c]
        time.sleep(0.05)
        return None


class KeyQueue:
    def __init__(
        self,
        clear_key: str = "C",
        digit_key: str = "D",
        eol_key: str = "#",
        swap_key: str = "*",
        max_length: int = 256,
    ) -> None:
        self._deque: deque[str] = deque(maxlen=max_length)
        self._clear_key = clear_key
        self._digit_key = digit_key
        self._eol_key = eol_key
        self._cv = Condition()
        self._excluded_keys = {k for k in [clear_key, digit_key, eol_key, swap_key] if k is not None}
        self._keypress_ev = Event()
        self._eol_ev = Event() if eol_key else None
        self._clear_ev = Event() if clear_key else None
        self._digit_ev = Event() if digit_key else None

    def __len__(self) -> int:
        return len(self.key_presses)

    def keypress_handler(self) -> Callable:
        def fn(keypad: Keypad) -> None:
            keypress = keypad.keypress
            if keypress:
                with self._cv:
                    # don't clear digit_ev, we need it for context
                    for ev in [self._keypress_ev, self._eol_ev, self._clear_ev]:
                        if ev:
                            ev.clear()
                    if keypress == self._clear_key:
                        self._deque.clear()
                        self._clear_ev.set()
                    elif keypress == self._digit_key:
                        self._deque.clear()
                        self._digit_ev.set()
                    elif keypress == self._eol_key:
                        self._eol_ev.set()
                    else:
                        self._deque.extend(keypress)
                    self._keypress_ev.set()
                    self._cv.notify_all()

        return fn

    __call__ = keypress_handler

    @property
    def key_presses(self) -> str:
        with self._cv:
            return "".join([c for c in self._deque if c not in self._excluded_keys])

    def processed_digit(self) -> None:
        with self._cv:
            if self._digit_ev and self._digit_ev.is_set():
                self._digit_ev.clear()
                self._deque.clear()

    def reset(self) -> None:
        with self._cv:
            self._keypress_ev.set()  # force controllers to wake up
            for ev in [self._eol_ev, self._clear_ev, self._digit_ev, self._keypress_ev]:
                if ev:
                    ev.clear()
            self._deque.clear()
            self._cv.notify_all()

    @property
    def is_eol(self) -> bool:
        return self._eol_ev.is_set() if self._eol_ev else False

    @property
    def is_clear(self) -> bool:
        return self._clear_ev.is_set() if self._clear_ev else False

    @property
    def is_digit(self) -> bool:
        return self._digit_ev.is_set() if self._digit_ev else False

    def wait_for_keypress(self, timeout: float = 10) -> str | None:
        try:
            self._keypress_ev.wait(timeout)
            if (
                self._keypress_ev.is_set() is False
                or (self._eol_ev and self._eol_ev.is_set())
                or (self._clear_ev and self._clear_ev.is_set())
                or (self._digit_ev and self._digit_ev.is_set() and len(self._deque) == 0)
            ):
                return None
            else:
                return self._deque[-1] if len(self._deque) > 0 else None
        finally:
            self._keypress_ev.clear()

    def wait_for_eol(self, timeout: float = 10) -> str | None:
        if self._eol_ev:
            self._eol_ev.wait(timeout)
            if self._eol_ev.is_set():
                self._eol_ev.clear()
                return self.key_presses
        return None



from __future__ import annotations

import logging
import socket
import threading
from collections import deque, defaultdict
from queue import Queue
from threading import Thread
from typing import Tuple, Generic

from .base_req import BaseReq
from .constants import PDI_SOP, PDI_STF, PDI_EOP, PdiAction, PdiCommand
from .pdi_req import PdiReq, TmccReq
from ..comm.command_listener import Topic, Message, Channel, Subscriber, CommandDispatcher, SYNC_COMPLETE
from ..comm.enqueue_proxy_requests import EnqueueProxyRequests
from ..protocol.constants import DEFAULT_QUEUE_SIZE, DEFAULT_BASE_PORT, BROADCAST_TOPIC, CommandScope, PROGRAM_NAME
from ..utils.ip_tools import get_ip_address

log = logging.getLogger(__name__)


class PdiListener(Thread):
    _instance: None = None
    _lock = threading.RLock()

    @classmethod
    def build(
        cls,
        base3: str = None,
        base3_port: int = DEFAULT_BASE_PORT,
        queue_size: int = DEFAULT_QUEUE_SIZE,
        build_base3_reader: bool = True,
    ) -> PdiListener:
        """
        Factory method to create a CommandListener instance
        """
        if base3 is None:
            build_base3_reader = False
        return PdiListener(base3, base3_port, queue_size, build_base3_reader)

    @classmethod
    def is_built(cls) -> bool:
        return cls._instance is not None

    @classmethod
    def is_running(cls) -> bool:
        # noinspection PyProtectedMember
        return cls._instance is not None and cls._instance._is_running is True

    @classmethod
    def stop(cls) -> None:
        with cls._lock:
            if cls._instance:
                cls._instance.shutdown()

    @classmethod
    def enqueue_command(cls, data: bytes | PdiReq) -> None:
        if cls._instance is not None and data:
            if isinstance(data, PdiReq):
                data = data.as_bytes
            # noinspection PyProtectedMember
            cls._instance._base3.send(data)

    @classmethod
    def listen_for(cls, listener: Subscriber, channel: Topic, address: int = None, action: PdiAction = None):
        if cls._instance is not None:
            cls._instance.dispatcher.subscribe(listener, channel, address, action)
        else:
            raise AttributeError("Pdi Listener not initialized")

    def __new__(cls, *args, **kwargs):
        """
        Provides singleton functionality. We only want one instance
        of this class in a process
        """
        with cls._lock:
            if PdiListener._instance is None:
                # noinspection PyTypeChecker
                PdiListener._instance = super(PdiListener, cls).__new__(cls)
                PdiListener._instance._initialized = False
            return PdiListener._instance

    def __init__(
        self,
        base3_addr: str,
        base3_port: int = DEFAULT_BASE_PORT,
        queue_size: int = DEFAULT_QUEUE_SIZE,
        build_base3_reader: bool = True,
    ) -> None:
        if self._initialized:
            return
        else:
            self._initialized = True
        self._base3_addr = base3_addr
        self._base3_port = base3_port
        super().__init__(daemon=True, name=f"{PROGRAM_NAME} PDI Listener {base3_addr}:{base3_port}")

        # open a connection to our Base 3
        if build_base3_reader is True:
            from .base3_buffer import Base3Buffer

            self._base3 = Base3Buffer(base3_addr, base3_port, queue_size, self)
        else:
            self._base3 = None

        # prep our consumer(s)
        self._cv = threading.Condition()
        self._deque = deque(maxlen=DEFAULT_QUEUE_SIZE)
        self._is_running = True
        self._dispatcher = PdiDispatcher.build(queue_size)

        # start listener thread
        self.start()

    @property
    def dispatcher(self) -> PdiDispatcher:
        return self._dispatcher

    def run(self) -> None:
        eop_pos = -1
        while self._is_running:
            # process bytes, as long as there are any
            with self._cv:
                if not self._deque:
                    self._cv.wait()  # wait to be notified
            # check if the first bite is in the list of allowable command prefixes
            dq_len = len(self._deque)
            while dq_len > 0 and self._is_running:  # may indicate thread is exiting
                # We now begin a state machine where we look for an SOP/EOP pair. Throw away
                # bytes until we see an SOP
                if self._deque[0] == PDI_SOP:
                    # We've found the possible start of a PDI command sequence. Check if we've found
                    # a PDI_EOP byte, or a "stuff" byte; we handle each situation separately
                    # if eop_pos != -1, we have to look past eop_pos for the next possible eop
                    try:
                        eop_pos = self._deque.index(PDI_EOP, eop_pos + 1)
                    except ValueError:
                        # no luck, wait for more bytes; should we impose a maximum byte count?
                        dq_len = -1  # to bypass inner while loop; we need more data
                        continue
                    # make sure the preceding byte isn't a stuff byte
                    if eop_pos - 1 > 0:
                        if self._deque[eop_pos - 1] == PDI_STF:
                            continue  # this EOP is part of the data stream; preceded by STF
                        # we found a complete PDI packet! Queue it for processing
                        req_bytes = bytes()
                        for _ in range(eop_pos + 1):
                            req_bytes += self._deque.popleft().to_bytes(1, byteorder="big")
                            dq_len -= 1
                        try:
                            if log.isEnabledFor(logging.DEBUG):
                                log.debug(f"Offering->0x{req_bytes.hex(':')}")
                            self._dispatcher.offer(PdiReq.from_bytes(req_bytes))
                        except Exception as e:
                            log.error(f"Failed to dispatch request: {req_bytes.hex(':')}")
                            log.exception(e)
                        finally:
                            eop_pos = -1
                        continue  # with while dq_len > 0 loop
                # pop this byte and continue; we either received unparsable input
                # or started receiving data mid-command
                log.warning(f"PdiListener Ignoring {hex(self._deque.popleft())}")
                dq_len -= 1
                eop_pos = -1
        # shut down the dispatcher
        if self._dispatcher:
            self._dispatcher.shutdown()

    def offer(self, data: bytes) -> None:
        if data:
            with self._cv:
                self._deque.extend(data)
                self._cv.notify()

    def shutdown(self) -> None:
        if hasattr(self, "_cv"):
            with self._cv:
                self._is_running = False
                self._cv.notify()
        if hasattr(self, "_dispatcher"):
            if self._dispatcher:
                self._dispatcher.shutdown()
        if hasattr(self, "_base3"):
            if self._base3:
                self._base3.shutdown()
        PdiListener._instance = None

    def subscribe(self, listener: Subscriber, channel: Topic, address: int = None, action: PdiAction = None) -> None:
        self._dispatcher.subscribe(listener, channel, address, action)

    def unsubscribe(self, listener: Subscriber, channel: Topic, address: int = None, action: PdiAction = None) -> None:
        self._dispatcher.unsubscribe(listener, channel, address, action)

    def subscribe_any(self, subscriber: Subscriber) -> None:
        self._dispatcher.subscribe_any(subscriber)

    def unsubscribe_any(self, subscriber: Subscriber) -> None:
        self._dispatcher.unsubscribe_any(subscriber)


class PdiDispatcher(Thread, Generic[Topic, Message]):
    """
    The PdiDispatcher thread receives parsed PdiReqs from the
    PdiListener and dispatches them to subscribing listeners
    """

    _instance = None
    _lock = threading.RLock()

    @classmethod
    def build(cls, queue_size: int = DEFAULT_QUEUE_SIZE) -> PdiDispatcher:
        """
        Factory method to create a CommandDispatcher instance
        """
        return PdiDispatcher(queue_size)

    @classmethod
    def is_built(cls) -> bool:
        return cls._instance is not None

    @classmethod
    def get(cls) -> PdiDispatcher:
        if cls._instance is None:
            raise AttributeError("PdiDispatcher has not been initialized")
        return cls._instance

    @classmethod
    def is_running(cls) -> bool:
        # noinspection PyProtectedMember
        return cls._instance is not None and cls._instance._is_running is True

    def __new__(cls, *args, **kwargs):
        """
        Provides singleton functionality. We only want one instance
        of this class in a process
        """
        with cls._lock:
            if PdiDispatcher._instance is None:
                PdiDispatcher._instance = super(PdiDispatcher, cls).__new__(cls)
                PdiDispatcher._instance._initialized = False
            return PdiDispatcher._instance

    def __init__(self, queue_size: int = DEFAULT_QUEUE_SIZE) -> None:
        if self._initialized:
            return
        else:
            self._initialized = True
        super().__init__(daemon=True, name=f"{PROGRAM_NAME} Pdi Dispatcher")
        self._channels: dict[Topic | Tuple[Topic, int], Channel[Message]] = defaultdict(Channel)
        self._cv = threading.Condition()
        self._is_running = True
        self._broadcasts = False
        self._queue = Queue[PdiReq](queue_size)
        self._tmcc_dispatcher = CommandDispatcher.build(queue_size)
        self._server_port = EnqueueProxyRequests.server_port() if EnqueueProxyRequests.is_built() else None
        self._server_ips = get_ip_address()
        self.start()

    @property
    def is_broadcasts_enabled(self) -> bool:
        return self._broadcasts

    def run(self) -> None:
        while self._is_running:
            with self._cv:
                if self._queue.empty():
                    self._cv.wait()
            if self._queue.empty():  # we need to do a second check in the event we're being shutdown
                continue
            cmd: PdiReq = self._queue.get()
            try:
                if log.isEnabledFor(logging.DEBUG):
                    log.debug(cmd)
                # update broadcast channels, mostly used for command echoing
                if self._broadcasts:
                    self.publish(BROADCAST_TOPIC, cmd)

                # publish dispatched pdi commands to listeners
                if isinstance(cmd, PdiReq):
                    if isinstance(cmd, BaseReq):
                        # on the PyTrain server, we need to know when the initial
                        # roster sync is complete; we do this by looking for the
                        # response to the request for info on Train 98...
                        if cmd.pdi_command == PdiCommand.BASE_TRAIN and cmd.tmcc_id == 98:
                            CommandDispatcher.get().offer(SYNC_COMPLETE)
                        if cmd.is_ack is True or cmd.is_active is False:
                            continue
                    # for TMCC requests, forward to TMCC Command Dispatcher
                    if isinstance(cmd, TmccReq):
                        self._tmcc_dispatcher.offer(cmd.tmcc_command, from_pdi=True)
                    elif (1 <= cmd.tmcc_id <= 9999) or (cmd.scope == CommandScope.BASE and cmd.tmcc_id == 0):
                        if hasattr(cmd, "action"):
                            self.publish((cmd.scope, cmd.tmcc_id, cmd.action), cmd)
                        self.publish((cmd.scope, cmd.tmcc_id), cmd)
                        self.publish(cmd.scope, cmd)

                        # Update clients of state change. Note that we DO NOT do this
                        # if the command is TMCC command received from the Base, as it
                        # has been handled via the call to tmcc_dispatcher.offer above
                        if self._server_port is not None:
                            self.update_client_state(cmd)
            except Exception as e:
                log.error(f"PdiDispatcher: Error publishing {cmd}")
                log.exception(e)
            finally:
                self._queue.task_done()

    # noinspection DuplicatedCode
    def update_client_state(self, command: PdiReq):
        """
        Update all PyTrain clients with the dispatched command. Used to keep
        client states in sync with serve
        """
        for client, port in EnqueueProxyRequests.clients():
            if client in self._server_ips and port == self._server_port:
                continue  # don't notify ourself
            try:
                with self._lock:
                    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                        s.connect((client, port))
                        s.sendall(command.as_bytes)
                        _ = s.recv(32)
            except ConnectionRefusedError:
                # ignore disconnects; client will receive state update on reconnect
                pass
            except Exception as e:
                log.warning(f"Exception while sending PDI state update {command} to {client}")
                log.exception(e)

    def offer(self, pdi_req: PdiReq | bytes) -> None:
        """
        Receive a command from the listener thread and dispatch it to subscribers.
        We do this in a separate thread so that the listener thread doesn't fall behind
        """
        try:
            if isinstance(pdi_req, bytes):
                pdi_req = PdiReq.from_bytes(pdi_req)
            if isinstance(pdi_req, PdiReq) and not pdi_req.is_ping and not pdi_req.is_ack:
                with self._cv:
                    self._queue.put(pdi_req)
                    self._cv.notify()  # wake up receiving thread
        except Exception as e:
            log.error(e)

    def shutdown(self) -> None:
        with self._cv:
            self._is_running = False
            self._cv.notify()
        PdiDispatcher._instance = None

    @staticmethod
    def _make_channel(channel: Topic, address: int = None, action: PdiAction = None) -> Topic | Tuple:
        if channel is None:
            raise ValueError("Channel required")
        elif address is None:
            return channel
        elif action is None:
            return channel, address
        else:
            return channel, address, action

    def publish(self, channel: Topic, message: Message) -> None:
        if channel in self._channels:  # otherwise, we would create a channel simply by referencing i
            self._channels[channel].publish(message)

    def subscribe(self, subscriber: Subscriber, channel: Topic, address: int = None, action: PdiAction = None) -> None:
        if channel == BROADCAST_TOPIC:
            self.subscribe_any(subscriber)
        else:
            self._channels[self._make_channel(channel, address, action)].subscribe(subscriber)

    def unsubscribe(
        self, subscriber: Subscriber, channel: Topic, address: int = None, command: PdiAction = None
    ) -> None:
        if channel == BROADCAST_TOPIC:
            self.unsubscribe_any(subscriber)
        else:
            channel = self._make_channel(channel, address, command)
            self._channels[channel].unsubscribe(subscriber)
            if len(self._channels[channel].subscribers) == 0:
                del self._channels[channel]

    def subscribe_any(self, subscriber: Subscriber) -> None:
        # receive broadcasts
        self._channels[BROADCAST_TOPIC].subscribe(subscriber)
        self._broadcasts = True

    def unsubscribe_any(self, subscriber: Subscriber) -> None:
        # receive broadcasts
        self._channels[BROADCAST_TOPIC].unsubscribe(subscriber)
        if not self._channels[BROADCAST_TOPIC].subscribers:
            self._broadcasts = False



#!/usr/bin/env python3

#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

#
import logging
from argparse import ArgumentParser
from typing import List

from ..protocol.multibyte.multibyte_constants import (
    TMCC2EffectsControl,
    TMCC2LightingControl,
    TMCC2MultiByteEnum,
    TMCC2RailSoundsDialogControl,
    TMCC2RailSoundsEffectsControl,
)
from ..protocol.sequence.sequence_constants import SequenceCommandEnum
from ..protocol.tmcc1.engine_cmd import EngineCmd as EngineCmdTMCC1
from ..protocol.tmcc1.tmcc1_constants import TMCC1EngineCommandEnum
from ..protocol.tmcc2.engine_cmd import EngineCmd as EngineCmdTMCC2
from ..protocol.tmcc2.tmcc2_constants import TMCC2EngineCommandEnum
from ..utils.argument_parser import PyTrainArgumentParser
from . import CliBase, CliBaseTMCC, DataAction

log = logging.getLogger(__name__)

AUX_COMMAND_MAP = {
    "on": "_ON",
    "off": "_OFF",
    "opt1": "_OPTION_ONE",
    "opt2": "_OPTION_TWO",
}


class EngineCli(CliBaseTMCC):
    @classmethod
    def command_parser(cls) -> ArgumentParser:
        engine_parser = PyTrainArgumentParser(add_help=False)
        engine_parser.add_argument("engine", metavar="Engine/Train", type=int, help="Engine/Train to operate")

        ops = engine_parser.add_mutually_exclusive_group()
        ops.add_argument(
            "-a", "-set_address", action="store_const", const="SET_ADDRESS", dest="option", help="Set engine address"
        )

        ops.add_argument("-aux1", dest="aux1", choices=["on", "off", "opt1", "opt2"], nargs="?", type=str, const="opt1")
        ops.add_argument("-aux2", dest="aux2", choices=["on", "off", "opt1", "opt2"], nargs="?", type=str, const="opt1")
        ops.add_argument("-aux3", dest="aux3", choices=["on", "off", "opt1", "opt2"], nargs="?", type=str, const="opt1")
        ops.add_argument(
            "-b", "-blow_horn", action="store_const", const="BLOW_HORN_ONE", dest="option", help="Blow horn"
        )
        ops.add_argument(
            "-bl",
            "-boost_level",
            action=DataAction,
            dest="option",
            choices=range(0, 8),
            metavar="0 - 7",
            type=int,
            nargs="?",
            default=3,
            const="BOOST_LEVEL",
            help="Boost level",
        )
        ops.add_argument(
            "-br", "-brake_speed", action="store_const", const="BRAKE_SPEED", dest="option", help="Boost speed"
        )
        ops.add_argument(
            "-bs", "-boost_speed", action="store_const", const="BOOST_SPEED", dest="option", help="Brake speed"
        )
        ops.add_argument(
            "-fc",
            "-front_coupler",
            action="store_const",
            const="FRONT_COUPLER",
            dest="option",
            help="Open front coupler",
        )
        ops.add_argument(
            "-fwd",
            "-forward_direction",
            action="store_const",
            const="FORWARD_DIRECTION",
            dest="option",
            help="Set forward direction",
        )
        ops.add_argument(
            "-hiss",
            "-cylinder_hiss",
            action="store_const",
            const="CYLINDER_HISS",
            dest="option",
            help="Cylinder cock hiss sound",
        )
        ops.add_argument(
            "-kl",
            "-brake_level",
            action=DataAction,
            dest="option",
            choices=range(0, 8),
            metavar="0 - 7",
            type=int,
            nargs="?",
            default=3,
            const="BRAKE_LEVEL",
            help="Brake level",
        )
        ops.add_argument(
            "-l",
            "-engine_labor",
            action=DataAction,
            dest="option",
            choices=range(0, 32),
            metavar="0 - 31",
            type=int,
            nargs="?",
            default=0,
            const="ENGINE_LABOR",
            help="Engine labor",
        )
        ops.add_argument(
            "-n",
            action=DataAction,
            dest="option",
            choices=range(0, 10),
            metavar="0 - 9",
            type=int,
            nargs="?",
            default=7,  # random radio chatter
            const="NUMERIC",
            help="Send numeric value",
        )
        ops.add_argument(
            "-pm", "-prime_mover", choices=["on", "off"], nargs="?", type=str, help="Prime mover sound on/off"
        )
        ops.add_argument(
            "-pop", "-pop_off", action="store_const", const="POP_OFF", dest="option", help="Pop off sounds"
        )
        ops.add_argument("-r", "-ring_bell", action="store_const", const="RING_BELL", dest="option", help="Ring bell")
        ops.add_argument("-reset", action="store_const", const="RESET", dest="option", help="Reset engine/train")
        ops.add_argument(
            "-rc", "-rear_coupler", action="store_const", const="REAR_COUPLER", dest="option", help="Open rear coupler"
        )
        ops.add_argument(
            "-rev",
            "-reverse_direction",
            action="store_const",
            const="REVERSE_DIRECTION",
            dest="option",
            help="Set reverse direction",
        )
        ops.add_argument(
            "-rpm",
            action=DataAction,
            dest="option",
            choices=range(0, 8),
            metavar="0 - 7",
            type=int,
            nargs="?",
            default=0,
            const="DIESEL_RPM",
            help="Diesel RPM Level",
        )
        ops.add_argument(
            "-s", "-stop_immediate", action="store_const", const="STOP_IMMEDIATE", dest="option", help="Stop immediate"
        )
        ops.add_argument("-stall", action="store_const", const="STALL", dest="option", help="Set stall")
        ops.add_argument(
            "-sdd",
            "-shutdown_delayed",
            action="store_const",
            const="SHUTDOWN_DELAYED",
            dest="option",
            help="Shutdown delayed with announcement",
        )
        ops.add_argument(
            "-sdi",
            "-shutdown_immediate",
            action="store_const",
            const="SHUTDOWN_IMMEDIATE",
            dest="option",
            help="Shutdown Immediate",
        )
        ops.add_argument("-sound", choices=["on", "off"], nargs="?", type=str, const="on", help="Sound on/off")
        ops.add_argument(
            "-sq", "-sequence_control", choices=["on", "off"], nargs="?", type=str, help="Sequence control on/off"
        )
        ops.add_argument(
            "-sud",
            "-start_up_delayed",
            action="store_const",
            const="START_UP_DELAYED",
            dest="option",
            help="Start up delayed (prime mover)",
        )
        ops.add_argument(
            "-sui",
            "-start_up_immediate",
            action="store_const",
            const="START_UP_IMMEDIATE",
            dest="option",
            help="Start up immediate",
        )
        ops.add_argument(
            "-t",
            "-toggle_direction",
            action="store_const",
            const="TOGGLE_DIRECTION",
            dest="option",
            help="Toggle direction",
        )
        ops.add_argument(
            "-tb",
            "-train_brake",
            action=DataAction,
            dest="option",
            choices=range(0, 8),
            metavar="0 - 7",
            type=int,
            nargs="?",
            default=1,
            const="TRAIN_BRAKE",
            help="Train brake",
        )
        ops.add_argument(
            "-v-", "-volume_down", action="store_const", const="VOLUME_DOWN", dest="option", help="Master volume down"
        )
        ops.add_argument(
            "-v+", "-volume_up", action="store_const", const="VOLUME_UP", dest="option", help="Master volume up"
        )

        # create subparsers to handle train/engine-specific operations
        sp = engine_parser.add_subparsers(dest="sub_command", help="Engine/train sub-commands")

        # Bell operations
        bell = sp.add_parser("bell", aliases=["be"], help="Bell operations", parent=engine_parser)
        bell_group = bell.add_mutually_exclusive_group()
        bell_group.add_argument(
            "-r",
            "-ring",
            action="store_const",
            const="RING_BELL",
            dest="option",
            default="RING_BELL",
            help="Ring bell",
        )
        bell_group.add_argument("-on", action="store_const", const="BELL_ON", dest="option", help="Turn bell on")
        bell_group.add_argument("-off", action="store_const", const="BELL_OFF", dest="option", help="Turn bell off")
        bell_group.add_argument(
            "-d",
            "-ding",
            action=DataAction,
            dest="option",
            choices=range(0, 4),
            metavar="0 - 3",
            type=int,
            nargs="?",
            default=3,
            const="BELL_ONE_SHOT_DING",
            help="Bell one shot ding",
        )
        bell_group.add_argument(
            "-s",
            "-slider",
            action=DataAction,
            dest="option",
            choices=range(2, 6),
            metavar="2 - 5",
            type=int,
            nargs="?",
            default=2,
            const="BELL_SLIDER_POSITION",
            help="Bell slider position",
        )

        # Horn operations
        horn = sp.add_parser("horn", aliases=["ho"], help="Horn operations", parent=engine_parser)
        horn_group = horn.add_mutually_exclusive_group()
        horn_group.add_argument(
            "-1",
            "-blow_horn_one",
            action="store_const",
            const="BLOW_HORN_ONE",
            dest="option",
            default="BLOW_HORN_ONE",
            help="Blow horn one",
        )
        horn_group.add_argument(
            "-2", "-blow_horn_two", action="store_const", const="BLOW_HORN_TWO", dest="option", help="Blow horn two"
        )
        horn_group.add_argument(
            "-grade_crossing",
            action="store_const",
            const="GRADE_CROSSING_SEQ",
            dest="option",
            help="Brade crossing sequence",
        )
        horn_group.add_argument(
            "-i",
            "-intensity",
            action=DataAction,
            dest="option",
            choices=range(0, 16),
            metavar="0 - 15",
            type=int,
            nargs="?",
            default=1,
            const="QUILLING_HORN",
            help="Quilling horn intensity",
        )

        # Momentum operations
        momentum = sp.add_parser("momentum", aliases=["mo"], help="Momentum operations", parent=engine_parser)
        mom_group = momentum.add_mutually_exclusive_group()
        mom_group.add_argument(
            "-low",
            action="store_const",
            const="MOMENTUM_LOW",
            dest="option",
            default="MOMENTUM_MEDIUM",  # strange, but default actions only applied if placed
            help="Set momentum to low",
        )  # on first argument of mutual group
        mom_group.add_argument(
            "-medium",
            action="store_const",
            const="MOMENTUM_MEDIUM",
            dest="option",
            help="Set momentum to medium",
        )
        mom_group.add_argument(
            "-high",
            "-maximum",
            "-x",
            action="store_const",
            const="MOMENTUM_HIGH",
            dest="option",
            help="Set momentum to high",
        )
        mom_group.add_argument("-off", action="store_const", const="MOMENTUM", dest="option", help="Set momentum off")

        mom_group.add_argument(
            "-absolute",
            action=DataAction,
            dest="option",
            choices=range(0, 8),
            metavar="0 - 7",
            type=int,
            nargs="?",
            default=0,
            const="MOMENTUM",
            help="Set absolute momentum",
        )

        # Smoke operations
        smoke = sp.add_parser("smoke", aliases=["sm"], help="Smoke operations", parent=engine_parser)
        smoke_group = smoke.add_mutually_exclusive_group()
        smoke_group.add_argument(
            "-low",
            action="store_const",
            const="SMOKE_LOW",
            dest="option",
            default="SMOKE_LOW",  # strange, but default actions only applied if placed
            help="Set smoke level to low",
        )  # on first argument of mutual group
        smoke_group.add_argument(
            "-medium",
            action="store_const",
            const="SMOKE_MEDIUM",
            dest="option",
            help="Set smoke level to medium",
        )
        smoke_group.add_argument(
            "-high",
            "-maximum",
            "-x",
            action="store_const",
            const="SMOKE_HIGH",
            dest="option",
            help="Set smoke level to high",
        )
        smoke_group.add_argument("-off", action="store_const", const="SMOKE_OFF", dest="option", help="Set smoke off")

        sound = sp.add_parser("sound", aliases=["so"], help="Sound operations", parent=engine_parser)
        sound_group = sound.add_mutually_exclusive_group()
        sound_group.add_argument("-auger", action="store_const", const="AUGER", dest="option", help="Auger sound")
        sound_group.add_argument(
            "-b", "-brake_squeal", action="store_const", const="BRAKE_SQUEAL", dest="option", help="Brake squeal sound"
        )
        sound_group.add_argument(
            "-r",
            "-brake_air_release",
            action="store_const",
            const="BRAKE_AIR_RELEASE",
            dest="option",
            help="Brake air release sound",
        )
        sound_group.add_argument(
            "-rpm",
            "-diesel_rpm",
            action=DataAction,
            dest="option",
            choices=range(0, 8),
            metavar="0 - 7",
            type=int,
            nargs="?",
            default=0,
            const="DIESEL_RPM",
            help="Diesel run level sound",
        )
        sound_group.add_argument(
            "-f",
            "-fu",
            "-fue",
            "-fuel",
            "-refueling",
            action="store_const",
            const="REFUELLING",
            dest="option",
            help="Refueling sound",
        )
        sound_group.add_argument(
            "-l", "-let_off", action="store_const", const="LET_OFF", dest="option", help="Short let-off sound"
        )
        sound_group.add_argument(
            "-ll",
            "-long_let_off",
            action="store_const",
            const="LET_OFF_LONG",
            dest="option",
            help="Long let-off sound",
        )
        sound_group.add_argument(
            "-water_injector",
            action="store_const",
            const="WATER_INJECTOR",
            dest="option",
            help="Water injector sound",
        )
        sound_group.add_argument(
            "-off", action="store_const", const="SOUND_OFF", dest="option", help="Turn all sounds off"
        )
        sound_group.add_argument(
            "-on", action="store_const", const="SOUND_ON", dest="option", help="Turn all sounds on"
        )

        # Speed operations
        sp_metavar = (
            "Engine/Train speed: 0 - 199 (Legacy) or 0 - 31 (TMCC) or stop, roll, restricted, "
            "slow, medium, limited, normal, or highball"
        )
        speed = sp.add_parser("speed", aliases=["sp"], help="Speed of engine/train", parent=engine_parser)
        speed.add_argument(
            "data", type=CliBase._validate_speed, action="store", metavar=sp_metavar, help="Absolute/Relative speed"
        )

        speed.add_argument(
            "-dialog",
            action="store_const",
            const="RAMPED_SPEED_DIALOG_SEQ",
            dest="option",
            help="Trigger tower/engineer dialog",
        )

        speed_group = speed.add_mutually_exclusive_group()
        speed_group.add_argument(
            "-absolute",
            action="store_const",
            const="RAMPED_SPEED_SEQ",
            dest="option",
            help="Set absolute speed (using ramp)",
        )
        speed_group.add_argument(
            "-immediate",
            action="store_const",
            const="ABSOLUTE_SPEED",
            dest="option",
            help="Set absolute speed (immediate)",
        )
        speed_group.add_argument(
            "-relative",
            action="store_const",
            const="RELATIVE_SPEED",
            dest="option",
            help="Set relative speed speed (-5 to 5)",
        )
        speed_group.set_defaults(option="RAMPED_SPEED_SEQ")

        # construct final parser with all components in order
        return PyTrainArgumentParser(
            "Control specified engine/train (1 - 99)",
            parents=[
                engine_parser,
                cls.multi_parser(),
                cls.train_parser(),
                cls.command_format_parser(),
                cls.cli_parser(),
            ],
        )

    def __init__(self, arg_parser: ArgumentParser = None, cmd_line: List[str] = None, do_fire: bool = True) -> None:
        super().__init__(arg_parser, cmd_line, do_fire)
        engine: int = self._args.engine
        option_data: int = self._args.data if "data" in self._args else 0
        try:
            option = self._decode_engine_option()  # raise ValueError if you can't decode
            if option is None:
                raise ValueError("Must specify an option, use -h for help")
            log.debug(self._args)
            scope = self._determine_scope()
            if self.is_tmcc2 or option.is_tmcc2:
                cmd = EngineCmdTMCC2(
                    engine, option, option_data, scope, baudrate=self._baudrate, port=self._port, server=self._server
                )
            else:
                cmd = EngineCmdTMCC1(
                    engine,
                    TMCC1EngineCommandEnum(option),
                    option_data,
                    scope,
                    baudrate=self._baudrate,
                    port=self._port,
                    server=self._server,
                )
            if self.do_fire:
                cmd.fire(
                    repeat=self._args.repeat,
                    delay=self._args.delay,
                    baudrate=self._baudrate,
                    port=self._port,
                    server=self._server,
                )
            self._command = cmd
        except ValueError as ve:
            log.exception(ve)

    def _decode_engine_option(self) -> TMCC1EngineCommandEnum | TMCC2EngineCommandEnum | TMCC2MultiByteEnum | None:
        """
        Decode the 'option' argument, if present, into a valid
        TMCC1EngineCommandDef, TMCC2EngineCommandDef, or one of the multiword TMCC2
        parameter enums. Use the specified command format, if present, to help resolve,
        as the enum classes share element names
        """
        if "option" not in self._args:
            return None

        # if option is None, check if an aux command was specified via
        # the aux1/aux2 arguments; only one should have a value
        option = self._args.option
        if not option or not option.strip():
            # construct the EngineOptionEnum by prepending the aux choice
            # (AUX1/AUX2) to the suffix based on the argument value
            # (on, off, opt1, opt2)
            if "aux1" in self._args and self._args.aux1 is not None:
                option = f"AUX1{AUX_COMMAND_MAP[self._args.aux1.lower()]}"
            elif "aux2" in self._args and self._args.aux2 is not None:
                option = f"AUX2{AUX_COMMAND_MAP[self._args.aux2.lower()]}"
            elif "aux3" in self._args and self._args.aux3 is not None:
                option = f"AUX3{AUX_COMMAND_MAP[self._args.aux3.lower()]}"
            elif "sound" in self._args and self._args.sound is not None:
                option = f"sound_{self._args.sound}".upper()
            elif "sequence_control" in self._args and self._args.sequence_control is not None:
                option = f"sequence_control_{self._args.sequence_control}".upper()
            elif "prime_mover" in self._args and self._args.prime_mover is not None:
                option = f"prime_mover{self._args.prime_mover}".upper()
            else:
                raise ValueError("Must specify an option, use -h for help")
        else:
            option = str(option).strip().upper()

        # reset option in args, for display purposes
        self._args.option = option

        # if scope is TMCC1, resolve via TMCC1EngineCommandDef
        if self.is_tmcc1:
            enum_classes = [TMCC1EngineCommandEnum]
        else:
            enum_classes = [
                TMCC2EngineCommandEnum,
                TMCC2RailSoundsDialogControl,
                TMCC2RailSoundsEffectsControl,
                TMCC2EffectsControl,
                TMCC2LightingControl,
                SequenceCommandEnum,
            ]
        for enum_class in enum_classes:
            if option in dir(enum_class):
                return enum_class[option]
        raise ValueError(f"Invalid {self.command_format.name} option: {option}")



from src.pytrain.gpio.i2c.i2c import I2C
from test.test_base import TestBase
from .smbusmock import MBusMock


class TestI2c(TestBase):
    def test_i2c_init(self) -> None:
        smbus = MBusMock()
        # noinspection PyTypeChecker
        i2c = I2C(smbus)
        assert i2c.bus == smbus

    def test_write_read_data(self) -> None:
        # noinspection PyTypeChecker
        i2c = I2C(MBusMock())
        assert i2c.read_from(0x20, 0x01) == 0xFF
        i2c.write_to(0x20, 0x01, 0x00)
        assert i2c.read_from(0x20, 0x01) == 0x00

    def test_read(self) -> None:
        # noinspection PyTypeChecker
        i2c = I2C(MBusMock())
        assert i2c.read(0x20) == 0xFF

    def test_scan(self) -> None:
        # noinspection PyTypeChecker
        i2c = I2C(MBusMock())
        devices = i2c.scan()
        assert len(devices) == 8
